import csv as csv_lib
from io import StringIO
import json


def _json_loads(value):
    try:
        return json.loads(value)
    except:
        return value


class Reports:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/reports'

    def list(self) -> list:
        """
        Return list of all built-in reports.

        :return: List of reports.
        """

        params = {
            'type': 'report'
        }
        return self.britive.get(self.base_url, params=params)

    def run(self, report_id: str, csv: bool = False, filter_expression: str = None) -> any:
        """
        Run a report.

        :param report_id: The ID of the report.
        :param csv: If True the result will be returned as a CSV string. If False (default) the result will be returned
            as a list where each time in the list is a dict representing the row of data.
        :param filter_expression: The filter to apply to the report. It is left to the caller to provide a syntactically
            correct filter expression string.
        :return: CSV string or list.
        """

        params = {}
        if filter_expression:
            params['filter'] = filter_expression
        csv_results = self.britive.get(f'{self.base_url}/{report_id}/csv', params=params)

        # convert csv to json - issue is that JSON response has max of 1k records returned so have to use CSV
        # as the base and convert to dict if the client asked for dict
        if csv:
            return csv_results
        else:
            dict_results = []
            for row in csv_lib.DictReader(StringIO(csv_results), quoting=csv_lib.QUOTE_MINIMAL):
                dict_results.append({k: _json_loads(v) for k, v in row.items()})
            return dict_results





from . import exceptions


class ApiTokens:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/token'

    def list(self) -> list:
        """
        Retrieve all API tokens in the system.

        :return: List of API tokens.
        """

        return self.britive.get(self.base_url)

    def get(self, token_id: str) -> dict:
        for token in self.list():
            if token['id'] == token_id:
                return token
        raise exceptions.ApiTokenNotFound()

    def create(self, name: str = None, expiration_days: int = 90) -> dict:
        """
        Create an API token with full administrator privileges.

        The token will only be returned in the response once. Future calls to `list` or `get` will not include the
        token value. Ensure the token returned from this method is stored or it will be unrecoverable.

        The token will be associated with the user identity that makes this API call.

        :param name: The name of the token.
        :param expiration_days: The number of days in which the token would expire since it was last used.
            Valid values are 1-90. Default is 90.
        :return: Details of the newly created token.
        """

        data = {
            'tokenExpirationDays': expiration_days
        }

        if name:
            data['name'] = name

        return self.britive.post(self.base_url, json=data)

    def revoke(self, token_id: str) -> None:
        """
        Revoke a token. Equivalent to deleting it.

        :param token_id: The ID of the token.
        :return: None
        """

        response = self.britive.delete(f'{self.base_url}/{token_id}')
        if response == 'Successfully revoked token':
            return None
        else:
            raise Exception(str(response))

    def delete(self, token_id: str) -> None:
        """
        Delete a token.

        Equivalent to revoking it. `delete` is provided for compatibility with other resource types.

        :param token_id: The ID of the token.
        :return: None
        """

        return self.revoke(token_id=token_id)

    def update(self, token_id: str, name: str = None, expiration_days: int = None) -> None:
        """
        Update a token.

        `name` and/or `expiration_days` needs to be provided. Both cannot be `None`.

        :param token_id: The ID of the token.
        :param name: The name of the token.
        :param expiration_days: The number of days in which the token would expire since it was last used.
            Valid values are 1-90. Default is 90.
        :return: None
        """

        if not name and not expiration_days:
            raise ValueError('name and/or expiration_days must be provided.')

        if not name and expiration_days:  # only updating expiration days which is a different API call
            data = {
                'tokenExpirationDays': expiration_days
            }
            return self.britive.patch(f'{self.base_url}/{token_id}', json=data)

        # updating the name and possibly the expiration days - if expiration days not provided make a call
        # to get the current expiration days and use that number.

        data = {
            'name': name,
            'tokenExpirationDays': expiration_days or self.get(token_id=token_id)['tokenExpirationDays']
        }
        return self.britive.put(f'{self.base_url}/{token_id}', json=data)







from .banner import SettingsBanner


# this class is just a logical grouping construct


class Settings:
    def __init__(self, britive):
        self.britive = britive
        self.banner = SettingsBanner(britive)




def validate_token_expiration(days):
    if not (1 <= days <= 90):
        raise ValueError(f'invalid token expiration value - must ust be between 1 and 90')


class ServiceIdentityTokens:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}'

    def create(self, service_identity_id: str, token_expiration_days: int = 90) -> dict:
        """
        Create a token for a given service identity.

        The token has the same privileges assigned to the service identity. When this token is created, the old token
        associated with the identity provider will be removed. A service identity can have only one token at any given
        time. The token that is generated will be returned only once.

        :param service_identity_id: The ID of the service identity.
        :param token_expiration_days: The number of days in which token would expire since it was last used.
            The token expiration days can be any value between 1 day and 90 days.
        :return:
        """

        validate_token_expiration(token_expiration_days)

        data = {
            'tokenExpirationDays': token_expiration_days
        }

        return self.britive.post(f'{self.base_url}/users/{service_identity_id}/tokens', json=data)

    def update(self, service_identity_id: str, token_expiration_days: int = 90) -> None:
        """
        Update the token expiration days for an existing service identity token.

        :param service_identity_id: The ID of the service identity.
        :param token_expiration_days: The number of days in which token would expire since it was last used.
            The token expiration days can be any value between 1 day and 90 days.
        :return: None
        """

        validate_token_expiration(token_expiration_days)

        data = {
            'tokenExpirationDays': token_expiration_days
        }

        return self.britive.post(f'{self.base_url}/users/{service_identity_id}/tokens', json=data)

    def get(self, service_identity_id: str) -> dict:
        """
        Return details of the token associated with the service identity. Only one token can exist per service identity.

        :param service_identity_id: The ID of the service identity.
        :return: Details of the service identity token
        """

        return self.britive.get(f'{self.base_url}/users/{service_identity_id}/tokens')




class NotificationMediums:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = (
            f'{self.britive.base_url}/v1/notification-service/notificationmediums'
        )

    def list(self) -> dict:
        """
        List all notification mediums.

        :return: List of all notification mediums
        """

        return self.britive.get(self.base_url)['result']

    def create(
        self,
        notification_medium_type: str,
        name: str,
        description: str = 'Default notification medium description',
        connection_parameters: dict = {}
    ) -> dict:
        """
        Create a new notification medium.

        :param notification_medium_type: the type of the notification medium
        :param name: the name of the notification medium
        :param description: the description of the notification medium
        :param connection_parameters: the connection parameters of the notification medium
                valid connection parameters:
                    URL : for slack, the URL
                    token : for slack, Auth Token for the Application BOT
                    Webhook URL : for teams, the URL of the teams webhook
        :return: Details of the newly created notification medium.
        """

        params = {
            'name': name,
            'description': description,
            'type': notification_medium_type,
            'connectionParameters': connection_parameters,
        }
        return self.britive.post(self.base_url, json=params)

    def get(self, notification_medium_id) -> dict:
        """
        Provide details of the given notification medium, from a notification medium id.

        :param notification_medium_id: The ID  of the notification medium.
        :return: Details of the specified notification medium.
        """

        return self.britive.get(f'{self.base_url}/{notification_medium_id}')

    def update(self, notification_medium_id: str, parameters: dict) -> dict:
        """
        Update a notification medium.

        :param notification_medium_id: The ID of the notification medium.
        :param parameters: Parameters to update, valid fields are...
            name: the name of the notification medium
            description: the description of the notification medium
            connectionParameters: the connection parameters of the notification medium
                valid connection parameters:
                    URL: for slack, the URL
                    token: for slack, Auth Token for the Application BOT
                    Webhook URL: for teams, the URL of the teams webhook
        :return: Details of the updated notification medium.
        """

        return self.britive.patch(
            f'{self.base_url}/{notification_medium_id}', json=parameters
        )

    def delete(self, notification_medium_id: str) -> None:
        """
        Deletes a notification medium.

        :param notification_medium_id: the ID of the notification medium
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{notification_medium_id}')

    def get_channels(self, notification_medium_id: str) -> dict:
        """
        Provide a list of all channels for the given notification medium (only for slack).

        :param notification_medium_id: The ID of the notification medium.
        :return: List of all channels for the given notification medium.
        """

        return self.britive.get(f'{self.base_url}/{notification_medium_id}/channels')




class CustomAttributes:
    def __init__(self, principal):
        self.britive = principal.britive
        self.base_url: str = principal.britive.base_url + '/users/{id}/custom-attributes'  # will .format(id=...) later

    def get(self, principal_id: str, as_dict: bool = False) -> any:
        """
        Retrieve the current custom attributes associated with the specified Service Identity or User.

        :param principal_id: The ID of the Service Identity or User.
        :param as_dict: Whether to return a key/value mapping vs the raw response which is a list.
        """
        response = self.britive.get(self.base_url.format(id=principal_id))
        if as_dict:
            attrs = {}
            for attribute in response:
                current_attr_id = attribute['attributeId']
                if current_attr_id in attrs.keys():  # need to handle the multi-value attributes
                    is_list = isinstance(attrs[attribute['attributeId']], list)

                    if not is_list:  # make it a list
                        attrs[current_attr_id] = [attrs[current_attr_id]]
                    attrs[current_attr_id].append(attribute['attributeValue'])
                else:
                    attrs[current_attr_id] = attribute['attributeValue']
            return attrs
        else:
            return response

    def add(self, principal_id: str, custom_attributes: dict) -> None:
        """
        Adds custom attribute mappings to the provided Service Identity or User.

        :param principal_id: The IDs of the Service Identity or User.
        :param custom_attributes: An attribute map where keys are the custom attribute ids or names and values are
            custom attribute values as strings or list of strings for multivalued attributes.
        """
        return self._modify(
            principal_id=principal_id,
            operation='add',
            custom_attributes=custom_attributes,
        )

    def remove(self, principal_id: str, custom_attributes: dict) -> None:
        """
        Removes custom attribute mapping from the provided Service Identity or User.

        :param principal_id: The IDs of the Service Identity or User.
        :param custom_attributes: An attribute map where keys are the custom attribute ids or names and values are
            custom attribute values as strings or list of strings for multivalued attributes.
        """
        return self._modify(
            principal_id=principal_id,
            operation='remove',
            custom_attributes=custom_attributes
        )

    def _build_list(self, operation: str, custom_attributes: dict):
        # first get list of existing custom identity attributes and build some helpers
        existing_attrs = [attr for attr in self.britive.identity_attributes.list() if not attr['builtIn']]
        existing_attr_ids = [attr['id'] for attr in existing_attrs]
        attrs_by_name = {attr['name']: attr['id'] for attr in existing_attrs}

        # for each custom_attribute key/value provided ensure we convert to ID and build the list
        attrs_list = []
        for id_or_name, value in custom_attributes.items():
            # obtain the custom attribute id
            custom_attribute_id = id_or_name
            if custom_attribute_id not in existing_attr_ids:
                custom_attribute_id = attrs_by_name.get(custom_attribute_id, None)
            if not custom_attribute_id:
                raise ValueError(f'custom identity attribute name {id_or_name} not found.')

            # and create the list dict entry for each value
            value = value if isinstance(value, list) else [value]  # handle multivalued attributes
            for v in value:
                attrs_list.append(
                    {
                        'op': operation,
                        'customUserAttribute': {
                            'attributeValue': v,
                            'attributeId': custom_attribute_id
                        }
                    }
                )
        return attrs_list

    def _modify(self, principal_id: str, operation: str, custom_attributes: dict) -> None:
        if operation not in ['add', 'remove']:
            raise ValueError('operation must either be add or remove')

        return self.britive.patch(
            self.base_url.format(id=principal_id),
            json=self._build_list(operation=operation, custom_attributes=custom_attributes)
        )



from .exceptions import \
    UserDoesNotHaveMFAEnabled, \
    UserNotAllowedToChangePassword, \
    UserNotAssociatedWithDefaultIdentityProvider
from .helpers.custom_attributes import CustomAttributes


valid_statues = ['active', 'inactive']


class Users:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/users'
        self.custom_attributes = CustomAttributes(self)

    def list(self, filter_expression: str = None) -> list:
        """
        Provide an optionally filtered list of all users.

        :param filter_expression: filter list of users based on name, status, or role. The supported operators
             are 'eq' and 'co'. Example: 'name co "Smith"'
        :return: List of user records
        """

        params = {
            'type': 'User',
            'page': 0,
            'size': 100
        }
        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(self.base_url, params)

    def get(self, user_id: str) -> dict:
        """
        Provide details of the given user.

        :param user_id: The ID  of the user.
        :return: Details of the specified user.
        """

        return self.britive.get(f'{self.base_url}/{user_id}')

    def get_by_name(self, name: str) -> list:
        """
        Return the list of users whose name contains `name`.

        :param name: The name (or part of the name) of the user you wish to get
        :return: Details of the specified users. If no user is found will return an empty list.
        """

        return self.list(filter_expression=f'name co "{name}"')

    def get_by_status(self, status: str) -> list:
        """
        Return a list of users filtered to `status`.

        :param status: valid values are `active` and `inactive`
        :return: List of users.
        """

        if status not in valid_statues:
            raise ValueError(f'status {status} not allowed.')

        return self.list(filter_expression=f'status eq "{status}"')

    def search(self, search_string: str) -> list:
        """
        Search all user fields for the given `search_string`.

        :param search_string: String to search.
        :return: List of user records.
        """

        params = {
            'type': 'User',
            'page': 0,
            'size': 100,
            'searchText': search_string
        }

        return self.britive.get(self.base_url, params)

    def create(self, idp: str = None, **kwargs) -> dict:
        """
        Create a new user record.

        :param idp: The ID of the IdP with which to associate the user. If not specified will default to
            the Britive identity provider.
        :param kwargs: Valid fields are...
            email - required
            username - required
            password - required only if no `idp` is provided, otherwise the IdP will handle authentication
            lastName - required
            firstName - required
            mobile
            phone
            status - valid values are active, inactive - if omitted then will default to active
            adminRoles - in format [{'name': 'TenantAdmin},]
        :return: Details of the newly created user.
        """

        required_fields = [
            'email',
            'username',
            'lastName',
            'firstName',
            'status'
        ]

        kwargs['type'] = 'User'
        if idp:
            kwargs['identityProvider'] = {
                'id': idp
            }
        else:
            required_fields.append('password')

        if 'status' not in kwargs.keys():
            kwargs['status'] = 'active'

        if kwargs['status'] not in valid_statues:
            raise ValueError(f'invalid status {kwargs["status"]}')

        if not all(x in kwargs.keys() for x in required_fields):
            raise ValueError('Not all required keyword arguments were provided.')

        response = self.britive.post(self.base_url, json=kwargs)
        return response

    def update(self, user_id: str, **kwargs) -> dict:
        """
        Update the specified attributes of the provided user.

        :param user_id: The ID of the user to update
        :param kwargs: The attributes to update for the user
        :return: A dict containing the newly updated user details
        """

        # we first have to get the existing user so we can inject the username and email into the update request
        user = self.get(user_id)

        # add some required elements to the kwargs passed in by the caller
        kwargs['type'] = 'User'

        if 'username' not in kwargs.keys():
            kwargs['username'] = user['username']

        if 'email' not in kwargs.keys():
            kwargs['email'] = user['email']

        self.britive.patch(f'{self.base_url}/{user_id}', json=kwargs)

        # return the updated user record
        return self.get(user_id)

    def delete(self, user_id: str) -> None:
        """
        Delete a user.

        :param user_id: The ID of the user to update
        :return: None
        """

        self.britive.delete(f'{self.base_url}/{user_id}')

    def enable(self, user_id: str = None, user_ids: list = None) -> object:
        """
        Enable the given user(s).

        You can pass in both `user_id` for a single user and `user_ids` to enable multiple users in one call. If both
        `user_id` and `user_ids` are provided they will be merged together into one list.

        :param user_id: The ID of the user you wish to enable.
        :param user_ids: A list of user IDs that you wish to enable.
        :return: if `user_ids` is set will return a list of user records, else returns a user dict
        """

        computed_users = []
        if user_ids:
            computed_users += user_ids
        if user_id:
            computed_users.append(user_id)

        # de-dup
        computed_users = list(set(computed_users))
        response = self.britive.post(f'{self.base_url}/enabled-statuses', json=computed_users)
        if not user_ids:
            return response[0]
        return response

    def disable(self, user_id: str = None, user_ids: list = None) -> object:
        """
        Disable the given user(s).

        You can pass in both `user_id` for a single user and `user_ids` to disable multiple users at in one call.
        If both `user_id` and `user_ids` are provided they will be merged together into one list.

        :param user_id: The ID of the user you wish to disable.
        :param user_ids: A list of user IDs that you wish to disable.
        :return: if `user_ids` is set will return a list of user records, else returns a user dict
        """

        computed_users = []
        if user_ids:
            computed_users += user_ids
        if user_id:
            computed_users.append(user_id)

        # de-dup
        computed_users = list(set(computed_users))
        response = self.britive.post(f'{self.base_url}/disabled-statuses', json=computed_users)
        if not user_ids:
            return response[0]
        return response

    def reset_password(self, user_id: str, password: str) -> None:
        """
        Reset a user's password.

        Only applicable for users using the default Britive IdP.

        :param user_id: The ID of the user for which the password will be reset.
        :param password: The new password. Must match the password policy of
            at least 8 characters
            at least 1 uppercase character
            at least 1 lowercase character
            at least 1 digit
            at least 1 symbol
        :return: None
        """

        user = self.get(user_id)
        if not user['canChangeOrResetPassword']:
            raise UserNotAllowedToChangePassword()

        if user['identityProvider']['type'] != 'DEFAULT':
            raise UserNotAssociatedWithDefaultIdentityProvider()

        return self.britive.post(f'{self.base_url}/{user_id}/resetpassword', json={'password': password})

    def reset_mfa(self, user_id: str) -> None:
        """
        Reset a user's MFA.

        The user will be prompted to re-establish MFA at next login.

        :param user_id: The ID of the user for which MFA will be reset.
        :return: None
        """

        user = self.get(user_id)
        if not user['identityProvider']['mfaEnabled']:
            raise UserDoesNotHaveMFAEnabled()

        return self.britive.patch(f'{self.base_url}/{user_id}/resetmfa')

    def minimized_user_details(self, user_id: str = None, user_ids: list = []) -> list:
        """
        Retrieve a small set of user fields given a user id.

        :param user_id: The ID of the user. Will be combined with `user_ids`.
        :param user_ids: The list of user ids. Will be combined with `user_id`.
        :return: List of users with a small set of attributes.
        """
        if user_ids is None:
            user_ids = []
        if user_id and user_id not in user_ids:
            user_ids.append(user_id)
        if len(user_ids) == 0:
            return []

        return self.britive.post(f'{self.base_url}/minimized-user-details', json=user_ids)



import json

from .cache import *  # will also import some globals like `britive`


# starting with map so we can get a cached account to use for testing
@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_map(cached_application, cached_environment, cached_user, cached_account):
    response = britive.accounts.map(
        user_id=cached_user['userId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id'],
        account_id=cached_account['accountId']
    )
    assert isinstance(response, list)
    assert len(response) > 0
    assert cached_user['userId'] in [u['userId'] for u in response]


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_mapped_users(cached_application, cached_environment, cached_user, cached_account):
    response = britive.accounts.mapped_users(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id'],
        account_id=cached_account['accountId']
    )
    assert isinstance(response, list)
    assert len(response) > 0
    assert cached_user['userId'] in [u['userId'] for u in response]


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_users_available_to_map(cached_application, cached_environment, cached_user, cached_account):
    response = britive.accounts.users_available_to_map(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id'],
        account_id=cached_account['accountId']
    )
    assert isinstance(response, list)
    assert len(response) > 0
    assert cached_user['userId'] not in [u['userId'] for u in response]


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_unmap(cached_application, cached_environment, cached_user, cached_account):
    response = britive.accounts.unmap(
        user_id=cached_user['userId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id'],
        account_id=cached_account['accountId']
    )
    assert isinstance(response, list)
    assert cached_user['userId'] not in [u['userId'] for u in response]


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_list(cached_application, cached_environment):
    accounts = britive.accounts.list(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(accounts, list)
    assert len(accounts) > 0
    assert isinstance(accounts[0], dict)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_permissions(cached_application, cached_environment, cached_account):
    permissions = britive.accounts.permissions(
        account_id=cached_account['accountId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )

    assert isinstance(permissions, list)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_groups(cached_application, cached_environment, cached_account):
    groups = britive.accounts.groups(
        account_id=cached_account['accountId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )

    assert isinstance(groups, list)




import time
from .cache import *  # will also import some globals like `britive`


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_scan(cached_scan):
    assert isinstance(cached_scan, dict)
    assert 'taskId' in cached_scan.keys()


# warning - this will take a while if a scan was just initiated! grab and coffee and come back later
@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_status(cached_scan):
    while True:
        status = britive.scans.status(task_id=cached_scan['taskId'])
        if status['status'] == 'Success':
            break
        if status['status'] == 'Error':
            break
        time.sleep(10)
    assert status['status'] == 'Success'


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_history(cached_application):
    response = britive.scans.history(application_id=cached_application['appContainerId'])
    assert isinstance(response, list)
    assert len(response) > 0


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_diff_accounts(cached_application, cached_environment):
    response = britive.scans.diff(
        resource='accounts',
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(response, list)
    assert len(response) > 0


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_diff_groups(cached_application, cached_environment):
    response = britive.scans.diff(
        resource='groups',
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(response, list)
    assert len(response) > 0


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_diff_permissions(cached_application, cached_environment):
    response = britive.scans.diff(
        resource='permissions',
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(response, list)
    assert len(response) > 0



from .cache import *  # will also import some globals like `britive`

user_keys = [
    'userId',
    'status',
    'external',
    'mappedAccounts',
    'identityProvider',
    'name',
    'firstName',
    'lastName',
    'username',
    'email',
    'type',
    'adminRoles'
]


def test_create(cached_user):
    assert isinstance(cached_user, dict)
    assert set(user_keys).issubset(cached_user.keys())


def test_list(cached_user):
    response = britive.users.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)
    assert set(user_keys).issubset(response[0].keys())
    assert cached_user['userId'] in [x['userId'] for x in response]


def test_get(cached_user):
    user = britive.users.get(cached_user['userId'])
    assert isinstance(user, dict)
    assert set(user_keys).issubset(user.keys())
    assert user['userId'] == cached_user['userId']


def test_get_by_name_co(cached_user):
    users = britive.users.get_by_name(cached_user['lastName'])
    assert isinstance(users, list)
    assert isinstance(users[0], dict)
    assert set(user_keys).issubset(users[0].keys())


def test_search(cached_user):
    users = britive.users.search(cached_user['email'].split('@')[0])
    assert isinstance(users, list)
    assert len(users) > 0
    assert isinstance(users[0], dict)
    assert set(user_keys).issubset(users[0].keys())


def test_get_by_status():
    users = britive.users.get_by_status('active')
    assert isinstance(users, list)
    assert len(users) > 0
    assert isinstance(users[0], dict)
    assert set(user_keys).issubset(users[0].keys())


def test_update(cached_user):
    user = britive.users.update(cached_user['userId'], mobile='1234567890')
    assert isinstance(user, dict)
    assert user['mobile'] == '1234567890'


def test_disable_single(cached_user):
    user = britive.users.disable(user_id=cached_user['userId'])
    assert isinstance(user, dict)
    assert user['status'] == 'inactive'


def test_enable_single(cached_user):
    user = britive.users.enable(user_id=cached_user['userId'])
    assert isinstance(user, dict)
    assert user['status'] == 'active'


def test_disable_list(cached_user):
    user = britive.users.disable(user_id=cached_user['userId'], user_ids=[cached_user['userId']])
    assert isinstance(user, list)
    assert user[0]['status'] == 'inactive'


def test_enable_list(cached_user):
    user = britive.users.enable(user_id=cached_user['userId'], user_ids=[cached_user['userId']])
    assert isinstance(user, list)
    assert user[0]['status'] == 'active'


def test_reset_password(cached_user):
    response = britive.users.reset_password(cached_user['userId'], generate_random_password())
    assert response is None


def test_reset_mfa(cached_user):
    with pytest.raises(exceptions.UserDoesNotHaveMFAEnabled):
        britive.users.reset_mfa(cached_user['userId'])


def test_set_custom_identity_attributes(cached_user, cached_identity_attribute):
    response = britive.service_identities.custom_attributes.add(
        principal_id=cached_user['userId'],
        custom_attributes={
            cached_identity_attribute['id']: [
                f'test-attr-value-{random.randint(0, 1000000)}'
            ]
        }
    )
    assert response is None


def test_get_custom_identity_attributes_list(cached_user, cached_identity_attribute):
    response = britive.service_identities.custom_attributes.get(
        principal_id=cached_user['userId'],
        as_dict=False
    )
    assert isinstance(response, list)
    assert len(response) == 1
    assert isinstance(response[0], dict)
    assert response[0]['attributeId'] == cached_identity_attribute['id']


def test_get_custom_identity_attributes_dict(cached_user, cached_identity_attribute):
    response = britive.service_identities.custom_attributes.get(
        principal_id=cached_user['userId'],
        as_dict=True
    )
    assert isinstance(response, dict)
    assert cached_identity_attribute['id'] in response.keys()
    assert response[cached_identity_attribute['id']].startswith('test-attr-value')


def test_remove_custom_identity_attributes(cached_user, cached_identity_attribute):
    value = britive.service_identities.custom_attributes.get(
        principal_id=cached_user['userId'],
        as_dict=True
    )[cached_identity_attribute['id']]
    response = britive.service_identities.custom_attributes.remove(
        principal_id=cached_user['userId'],
        custom_attributes={
            cached_identity_attribute['name']: [
                value
            ]
        }
    )
    assert response is None

    attributes = britive.service_identities.custom_attributes.get(
        principal_id=cached_user['userId'],
        as_dict=False
    )

    assert len(attributes) == 0


def test_minimized_user_details(cached_user):
    details = britive.users.minimized_user_details(user_id=cached_user['userId'])
    assert isinstance(details, list)
    assert len(details) == 1
    details = britive.users.minimized_user_details(user_ids=[cached_user['userId']])
    assert isinstance(details, list)
    assert len(details) == 1







from .cache import *  # will also import some globals like `britive`


def test_list():
    response = britive.system.roles.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)
    assert 'id' in response[0].keys()
    assert 'name' in response[0].keys()


def test_create(cached_system_level_role):
    assert isinstance(cached_system_level_role, dict)
    assert 'id' in cached_system_level_role.keys()
    assert 'name' in cached_system_level_role.keys()
    assert cached_system_level_role['name'].startswith('python-sdk')


def test_get_id(cached_system_level_role):
    response = britive.system.roles.get(role_identifier=cached_system_level_role['id'], identifier_type='id')
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_get_name(cached_system_level_role):
    response = britive.system.roles.get(role_identifier=cached_system_level_role['name'])
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_update_id(cached_system_level_role):
    permissions = britive.system.permissions.list()
    permission_ids = []
    permissions_to_add = ['SMAdminPermission', 'NMAdminPermission']
    for permission in permissions:
        if permission['name'] in permissions_to_add:
            permission_ids.append(permission['id'])

    role = britive.system.roles.build(
        name=cached_system_level_role['name'],
        permissions=permission_ids,
        identifier_type='id'
    )
    response = britive.system.roles.update(
        role_identifier=cached_system_level_role['id'],
        role=role,
        identifier_type='id'
    )

    assert response is None

    response = britive.system.roles.get(role_identifier=cached_system_level_role['id'], identifier_type='id')
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['permissions']) == 2


def test_update_name(cached_system_level_role):
    role = britive.system.roles.build(
        name=cached_system_level_role['name'],
        permissions=['NMAdminPermission', 'SMAdminPermission', 'ReportsViewPermission'],
        identifier_type='name'
    )
    response = britive.system.roles.update(
        role_identifier=cached_system_level_role['name'],
        role=role,
        identifier_type='name'
    )

    assert response is None

    response = britive.system.roles.get(role_identifier=cached_system_level_role['name'], identifier_type='name')
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['permissions']) == 3


def test_delete(cached_system_level_role):
    try:
        assert britive.system.roles.delete(role_identifier=cached_system_level_role['id'], identifier_type='id') is None

        with pytest.raises(exceptions.NotFound):
            britive.system.roles.get(cached_system_level_role['id'])
    finally:
        cleanup('role-system-level')




from . import exceptions
import time
from typing import Callable


approval_exceptions = {
    'rejected': exceptions.ProfileApprovalRejected(),
    'cancelled': exceptions.ProfileApprovalWithdrawn(),
    'timeout': exceptions.ProfileApprovalTimedOut()
}


class MyAccess:
    """
    This class is meant to be called by end users (as part of custom API integration work or the yet to be built
    Python based Britive CLI tooling). It is an API layer on top of the actions that can be performed on the
    "My Access" page of the Britive UI.
    
    No "administrative" access is required by the methods in this class. Each method will only return resources/allow
    actions which are permitted to be performed by the user/service identity, as identified by an API token or
    interactive login bearer token.

    It is entirely possible that an administrator who makes these API calls could get nothing returned, as that
    administrator may not have access to any profiles.
    """

    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/access'

    def list_profiles(self):
        """
        List the profiles for which the user has access.

        :return: List of profiles.
        """

        return self.britive.get(self.base_url)

    def list_checked_out_profiles(self) -> list:
        """
        Return list of details on currently checked out profiles for the user.

        :return: List of checked out profiles.
        """

        return self.britive.get(f'{self.base_url}/app-access-status')

    def get_checked_out_profile(self, transaction_id: str) -> dict:
        """
        Retrieve details of a given checked out profile.

        :param transaction_id: The ID of the transaction.
        :return: Details of the given profile/transaction.
        """

        for t in self.list_checked_out_profiles():
            if t['transactionId'] == transaction_id:
                return t
        raise exceptions.TransactionNotFound()

    def extend_checkout(self, transaction_id: str) -> dict:
        """
        Extend the expiration time of a currently checked out profile.

        :param transaction_id: The ID of the transaction.
        :return: Details of the given profile/transaction.
        """

        return self.britive.patch(f'{self.base_url}/extensions/{transaction_id}')

    def extend_checkout_by_name(self, profile_name: str, environment_name: str, application_name: str = None,
                                programmatic: bool = True) -> dict:
        """
        Extend the expiration time of a currently checked out profile by supplying the names of entities.

        :param profile_name: The name of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_name: The name of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param application_name: Optionally the name of the application, which can help disambiguate between profiles
            with the same name across applications.
        :param programmatic: True for programmatic credential checkout. False for console checkout.
        :return: Details of the given profile/transaction.
        """

        ids = self._get_profile_and_environment_ids_given_names(profile_name, environment_name, application_name)
        profile_id = ids['profile_id']
        environment_id = ids['environment_id']
        access_type = 'PROGRAMMATIC' if programmatic else 'CONSOLE'

        transaction_id = None
        for transaction in self.list_checked_out_profiles():
            is_profile = transaction['papId'] == profile_id
            is_environment = transaction['environmentId'] == environment_id
            is_type = transaction['accessType'] == access_type
            if all([is_profile, is_environment, is_type]):
                transaction_id = transaction['transactionId']
                break
        if not transaction_id:
            raise exceptions.TransactionNotFound()
        return self.extend_checkout(transaction_id=transaction_id)

    def request_approval_by_name(self, profile_name: str, environment_name: str, application_name: str = None,
                                 justification: str = None, wait_time: int = 60, max_wait_time: int = 600,
                                 block_until_disposition: bool = False, progress_func: Callable = None) -> any:
        """
        Requests approval to checkout a profile at a later time, using names of entities instead of IDs.

        Console vs. Programmatic access is not applicable here. The request for approval will allow the caller
        to checkout either type of access once the request has been approved.

        :param profile_name: The name of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_name: The name of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param application_name: Optionally the name of the application, which can help disambiguate between profiles
            with the same name across applications.
        :param justification: Optional justification if checking out the profile requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the profile checkout
            was approved. Only applicable if `block_until_disposition = True`.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception. Only applicable if `block_until_disposition = True`.
        :param block_until_disposition: Should this method wait/block until the request has been either approved,
            rejected, or withdrawn. If `True` then `wait_time` and `max_wait_time` will govern how long to wait before
            exiting.
        :param progress_func: An optional callback that will be invoked as the checkout process progresses.
        :return: If `block_until_disposition = True` then returns the final status of the request. If
            `block_until_disposition = False` then returns details about the approval request.
        :raises ProfileApprovalMaxBlockTimeExceeded: if max_wait_time has been reached while waiting for approval.
        """

        ids = self._get_profile_and_environment_ids_given_names(profile_name, environment_name, application_name)

        return self.request_approval(
            profile_id=ids['profile_id'],
            environment_id=ids['environment_id'],
            justification=justification,
            wait_time=wait_time,
            max_wait_time=max_wait_time,
            block_until_disposition=block_until_disposition,
            progress_func=progress_func
        )

    def request_approval(self, profile_id: str, environment_id: str, justification: str, wait_time: int = 60,
                         max_wait_time: int = 600, block_until_disposition: bool = False,
                         progress_func: Callable = None) -> any:
        """
        Requests approval to checkout a profile at a later time.

        Console vs. Programmatic access is not applicable here. The request for approval will allow the caller
        to checkout either type of access once the request has been approved.

        :param profile_id: The ID of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_id: The ID of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param justification: Optional justification if checking out the profile requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the profile checkout
            was approved. Only applicable if `block_until_disposition = True`.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception. Only applicable if `block_until_disposition = True`.
        :param block_until_disposition: Should this method wait/block until the request has been either approved,
            rejected, or withdrawn. If `True` then `wait_time` and `max_wait_time` will govern how long to wait before
            exiting.
        :param progress_func: An optional callback that will be invoked as the checkout process progresses.
        :return: If `block_until_disposition = True` then returns the final status of the request. If
            `block_until_disposition = False` then returns details about the approval request.
        :raises ProfileApprovalMaxBlockTimeExceeded: if max_wait_time has been reached while waiting for approval.
        """

        data = {
            'justification': justification
        }

        request = self.britive.post(
            f'{self.base_url}/{profile_id}/environments/{environment_id}/approvalRequest',
            json=data
        )

        if request is None:
            raise exceptions.ProfileCheckoutAlreadyApproved()

        request_id = request['requestId']

        if block_until_disposition:
            try:
                quit_time = time.time() + max_wait_time
                while True:
                    status = self.approval_request_status(request_id=request_id)['status'].lower()
                    if status == 'pending':
                        if time.time() >= quit_time:
                            raise exceptions.ProfileApprovalMaxBlockTimeExceeded()
                        if progress_func:
                            progress_func('awaiting approval')
                        time.sleep(wait_time)
                        continue
                    else:  # status == timeout or approved or rejected or cancelled
                        return status
            except KeyboardInterrupt:  # handle Ctrl+C (^C)
                # the first ^C we get we will try to withdraw the request
                # if we get another ^C while doing this we simply exit immediately
                try:
                    time.sleep(1)  # give the caller a small window to ^C again
                    self.withdraw_approval_request(request_id=request_id)
                    exit()
                except KeyboardInterrupt:
                    exit()

        else:
            return request

    def approval_request_status(self, request_id: str) -> dict:
        """
        Provides details on and approval request.

        :param request_id: The ID of the approval request.
        :return: Details of the approval request.
        """

        return self.britive.get(f'{self.britive.base_url}/v1/approvals/{request_id}')

    def withdraw_approval_request_by_name(self, profile_name: str, environment_name: str,
                                          application_name: str = None) -> None:
        """
        Withdraws a pending approval request, using names of entities instead of IDs.

        :param profile_name: The name of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_name: The name of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param application_name: Optionally the name of the application, which can help disambiguate between profiles
            with the same name across applications.
        :return: None
        """

        ids = self._get_profile_and_environment_ids_given_names(profile_name, environment_name, application_name)

        return self.withdraw_approval_request(
            profile_id=ids['profile_id'],
            environment_id=ids['environment_id']
        )

    def withdraw_approval_request(self, request_id: str = None, profile_id: str = None,
                                  environment_id: str = None) -> None:
        """
        Withdraws a pending approval request.

        Either `request_id` or (`profile_id` AND `environment_id`) are required.

        :param request_id: The ID of the approval request.
        :param profile_id: The ID of the profile.
        :param environment_id: The ID of the environment.
        :return: None
        """

        url = None
        if request_id:
            url = f'{self.britive.base_url}/v1/approvals/{request_id}'
        else:
            if not profile_id:
                raise ValueError('profile_id is required.')
            if not environment_id:
                raise ValueError('environment_id is required')
            url = f'{self.britive.base_url}/v1/approvals/consumer/papservice/resource?resourceId=' \
                  f'{profile_id}/{environment_id}'

        return self.britive.delete(url)

    def approve_request(self, request_id: str) -> None:
        """
        Approves a request.

        :param request_id: The ID of the request.
        :return: None.
        """

        params = {
            'approveRequest': 'yes'
        }

        return self.britive.patch(f'{self.britive.base_url}/v1/approvals/{request_id}', params=params)

    def reject_request(self, request_id: str):
        """
        Rejects a request.

        :param request_id: The ID of the request.
        :return: None.
        """

        params = {
            'approveRequest': 'no'
        }

        return self.britive.patch(f'{self.britive.base_url}/v1/approvals/{request_id}', params=params)

    def list_approvals(self) -> dict:
        """
        Lists approval requests.

        :return: List of approval requests.
        """

        params = {
            'requestType': 'myApprovals',
            'consumer': 'papservice'
        }

        return self.britive.get(f'{self.britive.base_url}/v1/approvals/', params=params)

    def _checkout(self, profile_id: str, environment_id: str, programmatic: bool = True,
                  include_credentials: bool = False, justification: str = None, wait_time: int = 60,
                  max_wait_time: int = 600, progress_func: Callable = None, iteration_num: int = 1) -> dict:

        params = {
            'accessType': 'PROGRAMMATIC' if programmatic else 'CONSOLE'
        }

        data = {
            'justification': justification
        }

        transaction = None

        # let's see if there is already a checked out profile
        progress_pending_checked_out_profiles_sent = False
        while True:  # will break the loop when needed
            loop = False
            if progress_func and not progress_pending_checked_out_profiles_sent:
                progress_func('reviewing currently checked out profiles')
                progress_pending_checked_out_profiles_sent = True
            for p in self.list_checked_out_profiles():
                right_profile = p['papId'] == profile_id
                right_env = p['environmentId'] == environment_id
                right_type = p['accessType'] == params['accessType']
                if all([right_profile, right_env, right_type]):
                    if p['checkedIn'] is None:  # still currently checked out so we can move on
                        transaction = p
                        break
                    else:  # we are in the middle of a profile being checked in so cannot check it out yet
                        loop = True
            if loop:
                if progress_func:
                    progress_func('pending profile checkin')
                time.sleep(1)
            else:
                break

        # if not check it out
        if not transaction:
            try:
                transaction = self.britive.post(
                    f'{self.base_url}/{profile_id}/environments/{environment_id}',
                    params=params,
                    json=data
                )
            except exceptions.InvalidRequest as e:
                if 'MA-0009' in str(e):  # old approval process that coupled approval and checkout
                    raise exceptions.ApprovalRequiredButNoJustificationProvided()
                elif 'MA-0010' in str(e):  # new approval process that de-couples approval from checkout
                    # if the caller has not provided a justification we know for sure the call will fail
                    # so raise the exception
                    if not justification:
                        raise exceptions.ApprovalRequiredButNoJustificationProvided()

                    # request approval
                    status = self.request_approval(
                        profile_id=profile_id,
                        environment_id=environment_id,
                        justification=justification,
                        wait_time=wait_time,
                        max_wait_time=max_wait_time,
                        block_until_disposition=True,
                        progress_func=progress_func
                    )

                    # handle the response based on the value of status
                    if status == 'approved':
                        transaction = self.britive.post(
                            f'{self.base_url}/{profile_id}/environments/{environment_id}',
                            params=params,
                            json=data
                        )
                    else:
                        raise approval_exceptions[status]
                elif 'e1001 - user has already checked out profile for this environment' in str(e).lower():
                    # this is a rare race condition...explained below
                    # if 2 or more calls from the same user to checkout a profile occur at the same time +/- 1/2 seconds
                    # both calls will get the list of checked out profiles and notice that the profile is not currently
                    # checked out (as the calls are made at the same time) and proceed to checkout the profile. One
                    # of the calls will "win" that race and the other call will get a http 400 response with the above
                    # error message. This is due to the first call checking out the profile and the second call making
                    # the api call to checkout the same profile...which has already been checked out. in this case we
                    # simply need to call this _checkout method again so it will realize the profile has already been
                    # checked out and poll for credentials. but we don't want to loop forever here in case some other
                    # issue is occurring so we will only allow 1 loop before raising the error
                    if iteration_num > 2:
                        raise e
                    return self._checkout(
                        profile_id=profile_id,
                        environment_id=environment_id,
                        programmatic=programmatic,
                        include_credentials=include_credentials,
                        justification=justification,
                        wait_time=wait_time,
                        max_wait_time=max_wait_time,
                        progress_func=progress_func,
                        iteration_num=iteration_num+1
                    )
                else:
                    raise e

        transaction_id = transaction['transactionId']

        # this approval workflow logic is for the legacy workflow when approval and checkout were coupled together
        # this logic can be removed once the new approval logic is deployed to production.
        if transaction['status'] == 'checkOutInApproval':  # wait for approval or until timeout occurs
            quit_time = time.time() + max_wait_time
            while True:
                try:
                    transaction = self.get_checked_out_profile(transaction_id=transaction_id)
                except exceptions.TransactionNotFound:
                    raise exceptions.ApprovalWorkflowRejected()
                if transaction['status'] == 'checkOutInApproval':  # we have an approval workflow occurring
                    if time.time() >= quit_time:
                        raise exceptions.ApprovalWorkflowTimedOut()
                    if progress_func:
                        progress_func('awaiting approval')
                    time.sleep(wait_time)
                    continue
                else:  # status == checkedOut
                    break

        # inject credentials if asked
        if include_credentials:
            # if the transaction is not in status of checkedOut here it will be after the
            # return of this call and we update the transaction object accordingly
            credentials, transaction = self.credentials(
                transaction_id=transaction_id,
                transaction=transaction,
                return_transaction_details=True,
                progress_func=progress_func
            )
            transaction['credentials'] = credentials

        if progress_func:
            progress_func('complete')
        return transaction

    def checkout(self, profile_id: str, environment_id: str, programmatic: bool = True,
                 include_credentials: bool = False, justification: str = None, wait_time: int = 60,
                 max_wait_time: int = 600, progress_func: Callable = None) -> dict:
        """
        Checkout a profile.

        If the profile has already been checked out this method will return the details of the checked out profile.

        If approval is required, this method will continue to check if approval has been obtained. Once the request
        is approved the profile will be checked out. Sending a `SIGINT/KeyboardInterrupt/Ctrl+C/^C` while waiting for
        the approval request to be dispositioned will withdraw the request. Sending a second `^C` immediately after
        the first will immediately exit the program.

        :param profile_id: The ID of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_id: The ID of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param programmatic: True for programmatic credential checkout. False for console checkout.
        :param include_credentials: True if tokens should be included in the response. False if the caller wishes to
            call `credentials()` at a later time. If True, the `credentials` key will be included in the response which
            contains the response from `credentials()`. Setting this parameter to `True` will result in a synchronous
            call vs. setting to `False` will allow for an async call.
        :param justification: Optional justification if checking out the profile requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the profile checkout
            was approved.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception.
        :param progress_func: An optional callback that will be invoked as the checkout process progresses.
        :return: Details about the checked out profile, and optionally the credentials generated by the checkout.
        :raises ApprovalRequiredButNoJustificationProvided: if approval is required but no justification is provided.
        :raises ApprovalWorkflowTimedOut: if max_wait_time has been reached while waiting for approval.
        :raises ApprovalWorkflowRejected: if the request to check out the profile was rejected.
        :raises ProfileApprovalTimedOut: if the approval request timed out exceeded the max time as specified by the
            profile policy.
        :raises ProfileApprovalRejected: if the approval request was rejected by the approver.
        :raises ProfileApprovalWithdrawn: if the approval request was withdrawn by the requester.
        """
        return self._checkout(
            profile_id=profile_id,
            environment_id=environment_id,
            programmatic=programmatic,
            include_credentials=include_credentials,
            justification=justification,
            wait_time=wait_time,
            max_wait_time=max_wait_time,
            progress_func=progress_func
        )

    def checkout_by_name(self, profile_name: str, environment_name: str, application_name: str = None,
                         programmatic: bool = True, include_credentials: bool = False, justification: str = None,
                         wait_time: int = 60, max_wait_time: int = 600, progress_func: Callable = None) -> dict:
        """
        Checkout a profile by supplying the names of entities vs. the IDs of those entities.

        If approval is required, this method will continue to check if approval has been obtained. Once the request
        is approved the profile will be checked out. Sending a `SIGINT/KeyboardInterrupt/Ctrl+C/^C` while waiting for
        the approval request to be dispositioned will withdraw the request. Sending a second `^C` immediately after
        the first will immediately exit the program.

        :param profile_name: The name of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_name: The name of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param application_name: Optionally the name of the application, which can help disambiguate between profiles
            with the same name across applications.
        :param programmatic: True for programmatic credential checkout. False for console checkout.
        :param include_credentials: True if tokens should be included in the response. False if the caller wishes to
            call `credentials()` at a later time. If True, the `credentials` key will be included in the response which
            contains the response from `credentials()`. Setting this parameter to `True` will result in a synchronous
            call vs. setting to `False` will allow for an async call.
        :param justification: Optional justification if checking out the profile requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the profile checkout
            was approved.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception.
        :param progress_func: An optional callback that will be invoked as the checkout process progresses.
        :return: Details about the checked out profile, and optionally the credentials generated by the checkout.
        :raises ApprovalRequiredButNoJustificationProvided: if approval is required but no justification is provided.
        :raises ApprovalWorkflowTimedOut: if max_wait_time has been reached while waiting for approval.
        :raises ApprovalWorkflowRejected: if the request to check out the profile was rejected.
        """

        ids = self._get_profile_and_environment_ids_given_names(profile_name, environment_name, application_name)

        return self._checkout(
            profile_id=ids['profile_id'],
            environment_id=ids['environment_id'],
            programmatic=programmatic,
            include_credentials=include_credentials,
            justification=justification,
            wait_time=wait_time,
            max_wait_time=max_wait_time,
            progress_func=progress_func
        )

    def credentials(self, transaction_id: str, transaction: dict = None, return_transaction_details: bool = False,
                    progress_func: Callable = None) -> any:
        """
        Return credentials of a checked out profile given the transaction ID.

        Will automatically determine the type of checkout (programmatic or console) and return the appropriate
        details.

        :param transaction_id: The ID of the transaction.
        :param transaction: Optional - the details of the transaction. Primary use is for internal purposes.
        :param return_transaction_details: Optional - whether to return the details of the transaction. Primary use is
            for internal purposes.
        :param progress_func: An optional callback that will be invoked as the checkout process progresses.
        :return: Credentials associated with the checked out profile represented by the specified transaction.
        """

        # step 1: get the details of the transaction so we can make the appropriate API call
        # we only need to get the details of the transaction if they are not already provided
        # or the transaction is not in the state of checkedOut
        if not transaction or transaction['status'] != 'checkedOut':
            while True:
                transaction = self.get_checked_out_profile(transaction_id=transaction_id)
                if transaction['status'] == 'checkOutSubmitted':  # async checkout process
                    if progress_func:
                        progress_func('credential creation')
                    time.sleep(1)
                    continue
                else:  # status == checkedOut
                    break

        # step 2: make the proper API call
        url_part = 'url' if transaction['accessType'] == 'CONSOLE' else 'tokens'
        creds = self.britive.get(f'{self.base_url}/{transaction_id}/{url_part}')

        if return_transaction_details:
            return creds, transaction
        else:
            return creds

    def checkin(self, transaction_id: str) -> dict:
        """
        Check in a checked out profile.

        :param transaction_id: The ID of the transaction.
        :return: Details of the checked in profile.
        """

        params = {
            'type': 'API'
        }
        return self.britive.put(f'{self.base_url}/{transaction_id}', params=params)

    def checkin_by_name(self, profile_name: str, environment_name: str, application_name: str = None) -> dict:
        """
        Check in a checked out profile by supplying the names of entities vs. the IDs of those entities

        :param profile_name: The name of the profile. Use `list_profiles()` to obtain the eligible profiles.
        :param environment_name: The name of the environment. Use `list_profiles()` to obtain the eligible environments.
        :param application_name: Optionally the name of the application, which can help disambiguate between profiles
            with the same name across applications.
        :return: Details of the checked in profile.
        """

        ids = self._get_profile_and_environment_ids_given_names(profile_name, environment_name, application_name)

        transaction_id = None
        for profile in self.list_checked_out_profiles():
            if profile['environmentId'] == ids['environment_id'] and profile['papId'] == ids['profile_id']:
                transaction_id = profile['transactionId']
                break
        if not transaction_id:
            raise ValueError(f'no checked out profile found for the given profile_name and environment_name')

        return self.checkin(transaction_id=transaction_id)

    def frequents(self) -> list:
        """
        Return list of frequently used profiles for the user.

        :return: List of profiles.
        """

        return self.britive.get(f'{self.base_url}/frequently-used')

    def favorites(self) -> list:
        """
        Return list of favorite profiles for the user.

        :return: List of profiles.
        """

        return self.britive.get(f'{self.base_url}/favorites')

    def whoami(self) -> dict:
        """
        Return details about the currently authenticated identity (user or service).

        :return: Details of the currently authenticated identity.
        """

        return self.britive.post(f'{self.britive.base_url}/auth/validate')['authenticationResult']

    def _get_profile_and_environment_ids_given_names(self, profile_name: str, environment_name: str,
                                                     application_name: str = None) -> dict:
        ids = None
        profile_found = False
        environment_found = False

        # collect relevant profile/environment combinations to which the identity is entitled
        for app in self.list_profiles():
            app_name = app['appName'].lower()
            if application_name and app_name != application_name.lower():  # restrict to one app if provided
                continue
            for profile in app['profiles']:
                prof_name = profile['profileName'].lower()
                prof_id = profile['profileId']

                if prof_name == profile_name.lower():
                    profile_found = True
                    for env in profile['environments']:
                        env_name = env['environmentName'].lower()
                        env_id = env['environmentId']

                        if env_name == environment_name.lower():
                            environment_found = True
                            # lets check to see if `ids` has already been set
                            # if so we should error because we don't know which name combo to use
                            if ids:
                                raise ValueError(
                                    f'multiple combinations of profile `{profile_name}` and environment '
                                    f'`{environment_name}` exist so no unique combination can be determined. Please '
                                    f'provide the optional parameter `application_name` to clarify which application '
                                    f'the environment belongs to.'
                                )
                            # set the IDs the first time
                            ids = {
                                'profile_id': prof_id,
                                'environment_id': env_id
                            }

        # do some error checking
        if not profile_found:
            raise ValueError(f'profile `{profile_name}` not found.')

        if profile_found and not environment_found:
            raise ValueError(f'profile `{profile_name}` found but not in environment `{environment_name}`.')

        # if we get here we found both the profile and environment and they are unique so
        # we can use the `ids` dict with confidence
        return ids




import datetime


class AuditLogs:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/logs'

    def fields(self) -> dict:
        """
        Return list of fields that be can used in a filter for an audit query.

        :return: Dict of field keys to field names.
        """

        return self.britive.get(f'{self.base_url}/fields')

    def operators(self) -> dict:
        """
        Return the list of operators that can be used in a filter for an audit query.

        :return: Dict of operator keys to operator names.
        """

        return self.britive.get(f'{self.base_url}/operators')

    def query(self, from_time: datetime = None, to_time: datetime = None, filter_expression: str = None,
              csv: bool = False) -> any:
        """
        Retrieve audit log events.

        `csv` options:

            - True: A CSV string is returned. The caller must persist the CSV string to disk.
            - False: A python list of audit events is returned.

        :param from_time: Lower end of the time frame to search. If not provided will default to
            7 days before `to_time`. `from_time` will be interpreted as if in UTC timezone so it is up to the caller to
            ensure that the datetime object represents UTC. No timezone manipulation will occur.
        :param to_time: Upper end of the time frame to search. If not provided will default to
            `datetime.datetime.utcnow()`. `to_time` will be interpreted as if in UTC timezone so it is up to the caller
            to ensure that the datetime object represents UTC. No timezone manipulation will occur.
        :param filter_expression: The expression used to filter the results. A list of available fields and operators
            can be found by querying `britive.audit_logs.fields()` and `britive.audit_logs.operators`, respectively.
            Multiple filter expressions must be joined together by `and`. No other join operator is support.
            Example: actor.displayName co "bob" and event.displayName eq "application"
        :param csv: Will result in a CSV string of the audit events being returned instead of a python list of events.
        :return: Either python list of events (dicts) or CSV string.
        :raises: ValueError - If from_time is greater than to_time.
        """

        to_time = to_time or datetime.datetime.utcnow()
        from_time = from_time or to_time - datetime.timedelta(days=7)

        if from_time > to_time:
            raise ValueError('from_time must occur before to_time.')

        params = {
            'from': from_time.isoformat(sep='T', timespec='seconds') + 'Z',
            'to': to_time.isoformat(sep='T', timespec='seconds') + 'Z'
        }
        if filter_expression:
            params['filter'] = filter_expression
        if not csv:
            params['size'] = 200

        return self.britive.get(f'{self.base_url}{"/csv" if csv else ""}', params=params)



from .cache import *  # will also import some globals like `britive`


def test_list():
    response = britive.system.policies.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)
    assert 'id' in response[0].keys()
    assert 'name' in response[0].keys()


def test_create(cached_system_level_policy):
    assert isinstance(cached_system_level_policy, dict)
    assert 'id' in cached_system_level_policy.keys()
    assert 'name' in cached_system_level_policy.keys()
    assert cached_system_level_policy['name'].startswith('python-sdk')


def test_create_default(cached_system_level_policy_condition_as_default_json_str):
    assert isinstance(cached_system_level_policy_condition_as_default_json_str, dict)
    assert 'id' in cached_system_level_policy_condition_as_default_json_str.keys()
    assert 'name' in cached_system_level_policy_condition_as_default_json_str.keys()
    assert cached_system_level_policy_condition_as_default_json_str['name'].startswith('python-sdk-')
    assert 'condition' in cached_system_level_policy_condition_as_default_json_str.keys()
    assert isinstance(cached_system_level_policy_condition_as_default_json_str['condition'], str)


def test_create_condition_dictionary(cached_system_level_policy_condition_as_dictionary):
    assert isinstance(cached_system_level_policy_condition_as_dictionary, dict)
    assert 'id' in cached_system_level_policy_condition_as_dictionary.keys()
    assert 'name' in cached_system_level_policy_condition_as_dictionary.keys()
    assert cached_system_level_policy_condition_as_dictionary['name'].startswith('python-sdk-')
    assert 'condition' in cached_system_level_policy_condition_as_dictionary.keys()
    assert isinstance(cached_system_level_policy_condition_as_dictionary['condition'], dict)


def test_create_single_nm(cached_tag):
    policy = britive.system.policies.build(
        name='python-sdk',
        tags=[cached_tag['name']],
        roles=['UserViewRole'],
        approval_notification_medium='Email',
        approver_tags=[cached_tag['name']]
    )

    assert isinstance(policy, dict)
    assert 'name' in policy.keys()
    assert policy['name'].startswith('python-sdk')


def test_create_multiple_nm(cached_tag):
    policy = britive.system.policies.build(
        name='python-sdk',
        tags=[cached_tag['name']],
        roles=['UserViewRole'],
        approval_notification_medium=['Email'],
        approver_tags=[cached_tag['name']]
    )

    assert isinstance(policy, dict)
    assert 'name' in policy.keys()
    assert policy['name'].startswith('python-sdk')


def test_get_id(cached_system_level_policy):
    response = britive.system.policies.get(policy_identifier=cached_system_level_policy['id'], identifier_type='id')
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_policies_condition_created_as_str_get_formatted_json(cached_system_level_policy_condition_as_default_json_str):
    response = britive.system.policies.get(
        policy_identifier=cached_system_level_policy_condition_as_default_json_str['id']
        , identifier_type='id'
        , condition_as_dict=False)
    assert 'condition' in response.keys()
    assert 'name' in response.keys()
    assert isinstance(response['condition'], str)


def test_policies_condition_created_as_str_get_formatted_dict(cached_system_level_policy_condition_as_default_json_str):
    response = britive.system.policies.get(
        policy_identifier=cached_system_level_policy_condition_as_default_json_str['id']
        , identifier_type='id'
        , condition_as_dict=True)
    assert 'condition' in response.keys()
    assert 'name' in response.keys()
    assert isinstance(response['condition'], dict)


def test_policies_condition_created_as_dict_get_formatted_json(cached_system_level_policy_condition_as_dictionary):
    response = britive.system.policies.get(policy_identifier=cached_system_level_policy_condition_as_dictionary['id']
                                           , identifier_type='id'
                                           , condition_as_dict=False)
    print(response)
    assert 'condition' in response.keys()
    assert 'name' in response.keys()
    assert isinstance(response['condition'], str)


def test_policies_condition_created_as_dict_get_formatted_dict(cached_system_level_policy_condition_as_dictionary):
    response = britive.system.policies.get(policy_identifier=cached_system_level_policy_condition_as_dictionary['id']
                                           , identifier_type='id'
                                           , condition_as_dict=True)
    assert 'condition' in response.keys()
    assert 'name' in response.keys()
    assert isinstance(response['condition'], dict)


def test_get_name(cached_system_level_policy):
    response = britive.system.policies.get(policy_identifier=cached_system_level_policy['name'])
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_update_id(cached_system_level_policy, cached_tag):
    roles = britive.system.roles.list()
    role_ids = []
    roles_to_add = ['ApplicationViewRole', 'TenantAdminRole']
    for role in roles:
        if role['name'] in roles_to_add:
            role_ids.append(role['id'])

    policy = britive.system.policies.build(
        name=cached_system_level_policy['name'],
        tags=[cached_tag['userTagId']],
        roles=role_ids,
        identifier_type='id'
    )
    response = britive.system.policies.update(
        policy_identifier=cached_system_level_policy['id'],
        policy=policy,
        identifier_type='id'
    )

    assert response is None

    response = britive.system.policies.get(policy_identifier=cached_system_level_policy['id'], identifier_type='id')
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['roles']) == 2


def test_update_name(cached_system_level_policy, cached_tag):
    policy = britive.system.policies.build(
        name=cached_system_level_policy['name'],
        tags=[cached_tag['name']],
        roles=['UserViewRole', 'NMAdminRole', 'ApplicationViewRole']
    )
    response = britive.system.policies.update(
        policy_identifier=cached_system_level_policy['name'],
        policy=policy,
        identifier_type='name'
    )

    assert response is None

    response = britive.system.policies.get(
        policy_identifier=cached_system_level_policy['name'],
        identifier_type='name'
    )
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['roles']) == 3


def test_disable_id(cached_system_level_policy):
    assert britive.system.policies.disable(
        policy_identifier=cached_system_level_policy['id'],
        identifier_type='id'
    ) is None
    assert britive.system.policies.get(
        policy_identifier=cached_system_level_policy['id'],
        identifier_type='id'
    )['isActive'] is False


def test_enable_id(cached_system_level_policy):
    assert britive.system.policies.enable(
        policy_identifier=cached_system_level_policy['id'],
        identifier_type='id'
    ) is None
    assert britive.system.policies.get(
        policy_identifier=cached_system_level_policy['id'],
        identifier_type='id'
    )['isActive'] is True


def test_disable_name(cached_system_level_policy):
    assert britive.system.policies.disable(policy_identifier=cached_system_level_policy['name']) is None
    assert britive.system.policies.get(
        policy_identifier=cached_system_level_policy['name']
    )['isActive'] is False


def test_enable_name(cached_system_level_policy):
    assert britive.system.policies.enable(policy_identifier=cached_system_level_policy['name']) is None
    assert britive.system.policies.get(
        policy_identifier=cached_system_level_policy['name']
    )['isActive'] is True


def test_delete(cached_system_level_policy):
    try:
        assert britive.system.policies.delete(
            policy_identifier=cached_system_level_policy['id'],
            identifier_type='id'
        ) is None
        with pytest.raises(exceptions.NotFound):
            britive.system.policies.get(cached_system_level_policy['id'])
    finally:
        cleanup('policy-system-level')


def test_evaluate():
    response = britive.system.policies.evaluate(
        statements=[
            {
                'action': 'authz.permission.update',
                'resource': '*',
                'consumer': 'authz'
            }
        ]
    )
    assert isinstance(response, dict)
    assert len(response.keys()) == 1
    assert 'PolicyEvalRequest' in list(response.keys())[0]




class Tasks:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/tasks'

    def list(self, task_service_id: str) -> list:
        """
        Return a list of tasks for the given `task_service_id`.

        Make a call to `britive.task_services.get()` to obtain the appropriate `task_service_id`.

        :param task_service_id: The ID of the task service.
        :return: List of tasks.
        """

        return self.britive.get(f'{self.base_url}/services/{task_service_id}/tasks')

    def get(self, task_service_id: str, task_id: str) -> dict:
        """
        Return details of a task.

        Make a call to `britive.task_services.get()` to obtain the appropriate `task_service_id`.

        :param task_service_id: The ID of the task service.
        :param task_id: The ID of the task.
        :return: Details of the task.
        """

        return self.britive.get(f'{self.base_url}/services/{task_service_id}/tasks/{task_id}')[0]

    def create(self, task_service_id: str, name: str, properties: dict, frequency_type: str, start_time: str = None,
               frequency_interval: str = None) -> dict:
        """
        Create a new task.

        Make a call to `britive.task_services.get()` to obtain the appropriate `task_service_id`.

        :param task_service_id: The ID of the task service.
        :param name: The name of the task.
        :param properties: This parameter is dependent on the task service type. Currently only `environmentScanner` is
            supported. Below are details about how to properly set this parameter for `environmentScanner`.
            * appId: this is the appContainerId of the app to be scanned.
            * scope: This is a list of scopes. Each scope can be of type EnvironmentGroup or Environment.
            * orgScan: Boolean indicating whether an org scan must be done or not.

            {
                "appId": "app-id",
                "scope": [
                    {
                        "type": "EnvironmentGroup"|"Environment",
                        "value": "ID"
                    }
                ],
                "orgScan": true|false
            }
        :param frequency_type: Valid values are Daily, Weekly, Monthly, Hourly (start_time is implicitly the next hour).
        :param start_time: The start time of the task in GMT. Only applies to Daily, Weekly, and Monthly
            frequency types. Example: `12:00`.
        :param frequency_interval: Only applies to Weekly, Monthly, and Hourly frequency types.
            * Weekly: possible values are 1-7, which is Mon-Sun, respectively.
            * Monthly: possible values are from 1-31.
            * Hourly: possible values are 1-23.
        :return: Details of the newly created task.
        """

        data = {
            'name': name,
            'startTime': start_time,
            'frequencyType': frequency_type,
            'frequencyInterval': frequency_interval,
            'properties': properties
        }
        return self.britive.post(f'{self.base_url}/services/{task_service_id}/tasks', json=data)

    def statuses(self, task_service_id: str, task_id: str) -> list:
        """
        Return a list of task statuses ordered by `sentTime` (part of the response) desc.

        :param task_service_id: The ID of the task service.
        :param task_id: The ID of the task.
        :return: List of task statuses.
        """

        return self.britive.get(f'{self.base_url}/services/{task_service_id}/tasks/{task_id}/statuses')

    def update(self, task_service_id: str, task_id: str, name: str = None, properties: dict = None,
               frequency_type: str = None, start_time: str = None, frequency_interval: str = None) -> dict:
        """
        Updates a task.

        Only provide parameters that should be updated.

        Make a call to `britive.task_services.get()` to obtain the appropriate `task_service_id`.

        :param task_service_id: The ID of the task service.
        :param task_id: The ID of the task.
        :param name: The name of the task.
        :param properties: This parameter is dependent on the task service type. Currently only `environmentScanner` is
            supported. Below are details about how to properly set this parameter for `environmentScanner`.
            * appId: this is the appContainerId of the app to be scanned.
            * scope: This is a list of scopes. Each scope can be of type EnvironmentGroup or Environment.
            * orgScan: Boolean indicating whether an org scan must be done or not.

            {
                "appId": "app-id",
                "scope": [
                    {
                        "type": "EnvironmentGroup"|"Environment",
                        "value": "ID"
                    }
                ],
                "orgScan": true|false
            }
        :param frequency_type: Valid values are Daily, Weekly, Monthly, Hourly (start_time is implicitly the next hour).
        :param start_time: The start time of the task in GMT. Only applies to Daily, Weekly, and Monthly
            frequency types. Example: `12:00`.
        :param frequency_interval: Only applies to Weekly, Monthly, and Hourly frequency types.
            * Weekly: possible values are 1-7, which is Mon-Sun, respectively.
            * Monthly: possible values are from 1-31.
            * Hourly: possible values are 1-23.
        :return: Details of the newly created task.
        """

        # construct the dict assuming every parameter is present
        raw_data = {
            'name': name,
            'startTime': start_time,
            'frequencyType': frequency_type,
            'frequencyInterval': frequency_interval,
            'properties': properties
        }

        # and now remove any None values
        data = {k: v for k, v in raw_data.items() if v is not None}

        return self.britive.patch(f'{self.base_url}/services/{task_service_id}/tasks/{task_id}', json=data)

    def delete(self, task_service_id: str, task_id: str) -> None:
        """
        Delete a task.

        :param task_service_id: The ID of the task service.
        :param task_id: The ID of the task.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/services/{task_service_id}/tasks/{task_id}')




from .system.policies import SystemPolicies
from typing import Union


class Policies:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/policies'

    @staticmethod
    def build(name: str, description: str = '', draft: bool = False, active: bool = True,
              read_only: bool = False, users: list = None, tags: list = None, tokens: list = None,
              service_identities: list = None, permissions: list = None, roles: list = None, ips: list = None,
              from_time: str = None, to_time: str = None, date_schedule: dict = None, days_schedule: dict = None,
              approval_notification_medium: Union[str, list] = None, time_to_approve: int = 5,
              access_validity_time: int = 120, approver_users: list = None, approver_tags: list = None,
              access_type: str = 'Allow', identifier_type: str = 'name', condition_as_dict: bool = False) -> dict:
        """
        Build a policy document given the provided inputs.

        :param name: The name of the policy.
        :param description: An optional description of the policy.
        :param draft: Indicates if the policy is a draft. Defaults to `False`.
        :param active: Indicates if the policy is active. Defaults to `True`.
        :param read_only: Indicates if the policy is a read only. Defaults to `False`.
        :param users: Optional list of user names or ids to which this policy applies.
        :param tags: Optional list of tag names or ids to which this policy applies.
        :param tokens: Optional list of token names or ids to which this policy applies.
        :param service_identities: Optional list of service identity names or ids to which this policy applies.
        :param permissions: Optional list of permission names or ids this policy grants. Provide either this parameter
            or `roles`.
        :param roles: Optional list of role names or ids to which this policy applies. Provider either this parameter
            or `permissions`.
        :param ips: Optional list of IP addresses for which this policy applies. Provide in CIDR notation
            or dotted decimal format for individual (/32) IP addresses.
        :param from_time: The start date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the start date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring start time. If this parameter is
            provided then `to_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param to_time: The end date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the end date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring end time. If this parameter is
            provided then `from_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param date_schedule: A dict in the format
            {
                'fromDate': '2022-10-29 10:30:00',
                'toDate': '2022-11-05 18:30:00',
                'timezone': 'UTC'
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param days_schedule: A dict in the format
            {
                'fromTime': '10:30:00',
                'toTime': '18:30:00',
                'timezone': 'UTC',
                'days': ['MONDAY','TUESDAY','WEDNESDAY','THURSDAY','FRIDAY','SATURDAY','SUNDAY']
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param approval_notification_medium: Optional notification medium name to which approval requests will be
            delivered. Can also specify a list of notification medium names. Specifying this parameter indicates the
            desire to enable approvals for this policy.
        :param time_to_approve: Optional number of minutes to wait for an approval before denying the action. Defaults
            to 5 minutes.
        :param access_validity_time: Optional number of minutes the access is valid after approval. Defaults to 120
            minutes.
        :param approver_users: Optional list of user names or ids who are to be considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param approver_tags: Optional list of tag names who are considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param access_type: The type of access this policy provides. Valid values are `Allow` and `Deny`. Defaults
            to `Allow`.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers are being provided to the other parameters. Either all identifiers must be names or all
            identifiers must be IDs.
        :param condition_as_dict: Prior to version 2.22.0 the only acceptable format for the condition block of
            a policy was as a stringifed json object. As of 2.22.0 the condition block can also be built as a raw
            python dictionary. This parameter will default to `False` to support backwards compatibility. Setting to
            `True` will result in the policy condition being returned/built as a python dictionary.
        :return: A dict which can be provided as a policy to `create` and `update`.
        """

        return SystemPolicies.build(
            name=name,
            description=description,
            draft=draft,
            active=active,
            read_only=read_only,
            users=users,
            tags=tags,
            tokens=tokens,
            service_identities=service_identities,
            permissions=permissions,
            roles=roles,
            ips=ips,
            from_time=from_time,
            to_time=to_time,
            date_schedule=date_schedule,
            days_schedule=days_schedule,
            approval_notification_medium=approval_notification_medium,
            time_to_approve=time_to_approve,
            access_validity_time=access_validity_time,
            approver_users=approver_users,
            approver_tags=approver_tags,
            access_type=access_type,
            identifier_type=identifier_type,
            condition_as_dict=condition_as_dict
        )






class Saml:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/saml'

    def settings(self, as_list: bool = False) -> any:
        """
        Retrieve the SAML settings for the tenant.

        For historical reasons there was a time in which multiple SAML settings could exist for a given tenant. This
        was to support certificate rotation. However a change in certificate issuance occurred and as a result this
        API call will only ever return 1 item in the list now. As such, `as_list` will default to `False` but returning
        the data as a list will continue to be supported for backwards compatibility.

        :param as_list: There is only 1 set of SAML settings per tenant. The API returns the settings
            as list but since there is only one this python library will return the single settings dict unless this
            parameter is set to True.
        :return: Details of the SAML settings for the tenant.
        """

        settings = self.britive.get(f'{self.base_url}/settings')
        if as_list:
            return settings
        return settings[0]

    def metadata(self) -> str:
        """
        Return the SAML metadata required in the SAML SSO configuration with a service provider.

        This operation is supported only in AWS and Oracle.

        The caller is responsible for saving this content to disk, if needed.

        :return: String representing the SAML XML metadata.
        """

        saml_id = self.settings(as_list=False)['id']

        return self.britive.get(f'{self.base_url}/metadata/{saml_id}')

    def certificate(self):
        """
        Return the SAML certificate required in the SAML SSO configuration with a service provider.

        This operation is applicable for applications that do not support importing SAML metadata.

        The caller is responsible for saving this content to disk, if needed.

        :return: String representing the SAML XML metadata.
        """

        saml_id = self.settings(as_list=False)['id']

        return self.britive.get(f'{self.base_url}/certificate/{saml_id}')



from .cache import *  # will also import some globals like `britive`


def test_environment_create(cached_environment):
    assert isinstance(cached_environment, dict)


def test_environment_update(cached_application, cached_environment):
    account_id = os.environ['BRITIVE_TEST_ENV_ACCOUNT_ID']
    response = britive.environments.update(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id'],
        accountId=account_id
    )
    assert isinstance(response, dict)
    assert 'catalogApplication' in response.keys()
    assert 'propertyTypes' in response['catalogApplication'].keys()
    account_id_response = None
    for prop in response['catalogApplication']['propertyTypes']:
        if prop['name'] == 'accountId':
            account_id_response = prop['value']
            break
    assert account_id_response == account_id


def test_environment_list_one(cached_application):
    envs = britive.environments.list(application_id=cached_application['appContainerId'])
    assert isinstance(envs, list)
    assert len(envs) == 1
    assert isinstance(envs[0], dict)


def test_environment_test(cached_application, cached_environment):
    response = britive.environments.test(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(response, dict)
    assert response['success']


def test_environment_get(cached_application, cached_environment):
    env = britive.environments.get(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(env, dict)
    assert env['environmentId'] == cached_environment['id']










class SystemRoles:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/roles'

    @staticmethod
    def _validate_identifier_type(identifier_type):
        if identifier_type not in ['id', 'name']:
            raise ValueError(f'identifier_type of {identifier_type} is invalid. Only `name` and `id` are allowed.')

    def list(self, filter_expression: str = '') -> list:
        """
        List system level roles.

        :param filter_expression: Filter based on `name`. Valid operators are `eq`, `sw`, and
            `co`. Example: name co security
        :returns: List of roles.
        """

        params = {}
        if filter_expression:
            params['filter'] = filter_expression
        return self.britive.get(self.base_url, params=params)

    def get(self, role_identifier: str, identifier_type: str = 'name', verbose: bool = False) -> dict:
        """
        Get details of the specified role.

        :param role_identifier: The ID or name of the role.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :param verbose: Whether to return a more compact response (the default) or a more verbose response.
        :returns: Details of the specified role.
        """

        self._validate_identifier_type(identifier_type)
        params = {
            'compactResponse': not verbose
        }
        return self.britive.get(f'{self.base_url}/{role_identifier}', params=params)

    def create(self, role: dict) -> dict:
        """
        Create a system level role.

        :param role: The role to create. Use `roles.build` to assist in constructing a proper role document.
        :returns: Details of the newly created role.
        """

        return self.britive.post(self.base_url, json=role)
    
    def update(self, role_identifier: str, role: dict, identifier_type: str = 'name') -> None:
        """
        Update a system level role.

        :param role_identifier: The ID or name of the role to update.
        :param role: The role to update. Use `roles.build` to assist in constructing a proper role document.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.patch(f'{self.base_url}/{role_identifier}', json=role)

    def delete(self, role_identifier: str, identifier_type: str = 'name') -> None:
        """
        Delete a system level role.

        :param role_identifier: The ID or name of the role to delete.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.delete(f'{self.base_url}/{role_identifier}')

    @staticmethod
    def build(name: str, permissions: list, description: str = '', read_only: bool = False,
              identifier_type: str = 'name') -> dict:
        """
        Build a role document given the provided inputs.

        :param name: The name of the role.
        :param permissions: List of permission names or ids this role grants.
        :param description: An optional description of the role.
        :param read_only: Indicates if the role is a read only. Defaults to `False`.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers are being provided to `permissions`. Either all identifiers must be names or all
            identifiers must be IDs.
        :return: A dict which can be provided as a role document to `create` and `update`.
        """

        # put it all together
        role = {
            'name': name,
            'description': description,
            'isReadOnly': read_only,
            'permissions': [{identifier_type: p} for p in permissions]
        }

        return role



class Environments:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def create(self, application_id: str, name: str, description: str = None, parent_group_id: str = None) -> dict:
        """
        Create a new environment inside the specified application.

        Only applicable to applications which have `catalogApplication.supportsEnvironmentScanning` set to `False`.

        :param application_id: The ID of the application inside of which the environment will be created.
        :param name: The name of the environment.
        :param description: An optional description of the environment.
        :param parent_group_id: An optional parent group id - if not provided then the root group is assumed and will
            be created as needed.
        :return: Details of the newly created environment.
        """

        data = {
            'name': name,
            'type': 'environment',
            'description': description or '',
            'parentGroupId': parent_group_id or self.britive.environment_groups.get_or_create_root(
                application_id=application_id
            )
        }

        return self.britive.post(f'{self.base_url}/{application_id}/root-environment-group/environments', json=data)

    def get(self, application_id: str, environment_id: str) -> dict:
        """
        Return details about the specified environment.

        :param application_id: The ID of the application where the environment resides.
        :param environment_id: The ID of the environment.
        :return: Details about the environment.
        """

        envs = self.list(application_id=application_id)
        for env in envs:
            if env['environmentId'] == environment_id:
                return env
        return {}

    def list(self, application_id: str) -> list:
        """
        Return a list of environments for the specified application.

        :param application_id: The ID of the application.
        :return: List of environments.
        """

        return self.britive.get(f'{self.base_url}/{application_id}/environments')

    def test(self, application_id: str, environment_id: str) -> dict:
        """
        Test the settings of the specified environment.

        :param application_id: The ID of the application containing the environment to test.
        :param environment_id: The ID fo the environment to test.
        :return: Results of the test. Failures will not throw errors - rather the failure message
            will be in the response.
        """

        return self.britive.post(f'{self.base_url}/{application_id}/environments/{environment_id}/tests')

    def update(self, application_id: str, environment_id: str,  **kwargs) -> dict:
        """
        Update the supplied properties of the environment.

        :param application_id: The ID of the application to update.
        :param environment_id: The ID of the environment to update.
        :param kwargs: A set of name/value pairs where the property name will be updated to the supplied value.
            Property names are dependent upon the catalog application. Call method `environment.get()` and review
            the attribute `catalogApplication.propertyTypes` for a list of acceptable names to use here.
        :return: Details of the updated environment.
        """

        data = {
            'propertyTypes': []
        }
        for key, value in kwargs.items():
            data['propertyTypes'].append(
                {
                    'name': key,
                    'value': value,
                    'defaultValue': value
                }
            )
        return self.britive.patch(
            f'{self.base_url}/{application_id}/environments/{environment_id}/properties',
            json=data
        )

    def scan(self, application_id: str, environment_id: str) -> dict:
        """
        Initiate a scan of the environment. Required only for applications which have
        `catalogApplication.supportsEnvironmentScanning` set to `False`.

        For all other application types the application will be scanned and not the underlying environment(s).

        Scans are asynchronous operations. The response will include a `taskId` which can be used to make calls
        to `britive.scans.status()` to obtain the current status of the scan.

        Note that scans can also be initiated from the Scans class. The same type of scan will be performed no matter
        where it is initiated.

        :param application_id: The ID of the application to scan.
        :param environment_id: The ID of the environment to scan.
        :return: Details of the scan that was initiated.
        """

        return self.britive.scans.scan(application_id=application_id, environment_id=environment_id)

    def delete(self, application_id: str, environment_id: str) -> None:
        """
        Delete the specified environment.

        :param application_id: The ID of the application containing the environment to delete.
        :param environment_id: The ID of the environment to delete.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{application_id}/environments/{environment_id}')






class TagMembershipRules:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/user-tags'

    def build(self, attribute_id_or_name: str, operator: str, value: str) -> dict:
        """
        Builds a membership rule.

        :param attribute_id_or_name: The attribute id or name for the rule. Names will be converted to ids by the SDK.
        :param operator: Valid values are `contains` and `is`.
        :param value: The value to match in the rule.
        :returns: Dictionary which can be used to build a list of rules.
        """

        if operator.lower() not in ['contains', 'is']:
            raise ValueError('invalid operator provided.')

        # first get list of existing identity attributes and build some helpers
        existing_attrs = [attr for attr in self.britive.identity_attributes.list()]
        existing_attr_ids = [attr['id'] for attr in existing_attrs]
        attrs_by_name = {attr['name']: attr['id'] for attr in existing_attrs}

        attribute_id = attribute_id_or_name
        if attribute_id not in existing_attr_ids:
            attribute_id = attrs_by_name.get(attribute_id_or_name, None)
        if not attribute_id:
            raise ValueError(f'identity attribute name {attribute_id_or_name} not found.')

        return {
            'attributeId': attribute_id,
            'operator': operator.lower(),
            'value': value
        }

    def list(self, tag_id: str) -> list:
        """
        List existing membership rules for the given tag.

        :param tag_id: The tag id.
        :returns: List of membership rules for the given tag.
        """

        return self.britive.get(f'{self.base_url}/{tag_id}/attribute-criteria')

    def create(self, tag_id: str, rules: list) -> list:
        """
        Create new membership rules for the given tag.

        :param tag_id: The tag id.
        :param rules: The list of rules to include in the request. Use `build` to help construct the list.
        :returns: List of newly created membership rules for the given tag.
        """

        return self.britive.post(f'{self.base_url}/{tag_id}/attribute-criteria', json=rules)

    def update(self, tag_id: str, rules: list) -> None:
        """
        Update membership rules for the given tag.

        All membership rules must be provided. What is provided in the request will fully replace the existing
        list of membership rules associated with the given tag.

        :param tag_id: The tag id.
        :param rules: The list of rules to include in the request. Use `build` to help construct the list.
        :returns: None.
        """

        return self.britive.patch(f'{self.base_url}/{tag_id}/attribute-criteria', json=rules)

    def delete(self, tag_id: str) -> None:
        """
        Delete all membership rules for the given tag.

        :param tag_id: The tag id.
        :returns: None.
        """

        return self.britive.patch(f'{self.base_url}/{tag_id}/attribute-criteria', json=[])

    def matched_users(self, tag_id: str) -> list:
        """
        Lists users which match the membership rules associated with the given tag.

        :param tag_id: The tag id.
        :returns: List of matching users.
        """

        params = {
            'page': 0,
            'size': 100
        }
        return self.britive.get(f'{self.base_url}/{tag_id}/matched-users', params=params)


class Tags:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/user-tags'
        self.membership_rules = TagMembershipRules(self.britive)

    def create(self, name: str, description: str = None, idp: str = None) -> dict:
        """
        Create a new tag.

        :param name: The tag name.
        :param description: The tag description.
        :param idp: The ID of the identity provider. To be used when creating an external tag.
        :return: Details of the newly created tag.
        """

        data = {
            'name': name,
            'description': description
        }

        if idp:
            data['userTagIdentityProviders'] = [{'identityProvider': {'id': idp}}]
            data['external'] = True

        return self.britive.post(self.base_url, json=data)

    def get(self, tag_id: str) -> dict:
        """
        Return details of a tag.

        :param tag_id: The ID of the tag.
        :return: Details of the tag.
        """

        return self.britive.get(f'{self.base_url}/{tag_id}')

    def list(self, filter_expression: str = None) -> list:
        """
        List all tags, optionally filtered via name or status.

        :param filter_expression: Filter the list of tags based on name or status. The supported operators are
            `eq' and `co`. An example is `status eq 'Active'`
        :return: List of tags.
        """

        params = {
            'page': 0,
            'size': 100
        }
        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(self.base_url, params)

    def search(self, search_string: str) -> list:
        """
        Searche all tag fields for the given `search_string` and returns
        a list of matched tags.

        :param search_string: String to search.
        :return: List of user records.
        """

        params = {
            'page': 0,
            'size': 100,
            'searchText': search_string
        }

        return self.britive.get(self.base_url, params)

    def users_for_tag(self, tag_id: str, filter_expression: str = None) -> list:
        """
         Retrieve the details of all users assigned to a tag,

        :param tag_id: The ID of the tag.
        :param filter_expression: Filter the list of users asssigned to `tag_id` based on name or status.
        The supported operators are`eq' and `co`. An example is `status eq 'Active'`
        :return: List of users associated with `tag_id`.
        """

        params = {
            'page': 0,
            'size': 100
        }
        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(f'{self.base_url}/{tag_id}/users', params)

    def available_users_for_tag(self, tag_id: str) -> list:
        """
        Retrieve all users available in the system that can possibly be added to a tag.

        :param tag_id: The ID of the tag.
        :return: List of users available to be assigned to `tag_id`.
        """

        return self.britive.get(f'{self.base_url}/{tag_id}/users?filter=available')

    def add_user(self, tag_id: str, user_id: str) -> dict:
        """
        Add a user to a tag.

        :param tag_id: The ID of the tag.
        :param user_id: The ID of the user.
        :return: Details of the user.
        """

        return self.britive.post(f'{self.base_url}/{tag_id}/users/{user_id}')

    def remove_user(self, tag_id: str, user_id: str) -> None:
        """
        Remove a user from a tag.

        :param tag_id: The ID of the tag.
        :param user_id: The ID of the user.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{tag_id}/users/{user_id}')

    def enable(self, tag_id: str) -> dict:
        """
        Enable a tag.

        :param tag_id: The ID of the tag.
        :return: Details of the tag.
        """

        return self.britive.post(f'{self.base_url}/{tag_id}/enabled-statuses')

    def disable(self, tag_id: str) -> dict:
        """
        Disable a tag.

        :param tag_id: The ID of the tag.
        :return: Details of the tag.
        """

        return self.britive.post(f'{self.base_url}/{tag_id}/disabled-statuses')

    def update(self, tag_id: str, name: str, description: str = None) -> dict:
        """
        Update the attributes of a tag.

        :param tag_id: The ID of the tag.
        :param name: The name of the tag.
        :param description: The description of the tag.
        :return: Details of the updated tag.
        """

        data = {
            'userTagId': tag_id,
            'name': name,
            'description': description
        }

        self.britive.patch(self.base_url, json=data)
        return self.get(tag_id)

    def delete(self, tag_id: str) -> None:
        """
        Delete a tag.

        You can delete an external tag ONLY if it has yet to be synced via SCIM.

        :param tag_id: The ID of the tag.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{tag_id}')

    def minimized_tag_details(self, tag_id: str = None, tag_ids: list = []) -> list:
        """
        Retrieve a small set of user fields given a user id.

        :param tag_id: The ID of the tag. Will be combined with `tag_ids`.
        :param tag_ids: The list of tag ids. Will be combined with `tag_id`.
        :return: List of tags with a small set of attributes.
        """
        if tag_ids is None:
            tag_ids = []
        if tag_id and tag_id not in tag_ids:
            tag_ids.append(tag_id)
        if len(tag_ids) == 0:
            return []

        return self.britive.post(f'{self.base_url}/minimized-tag-details', json=tag_ids)



import base64
import datetime
import hashlib
import hmac
import json
import os
from .. import exceptions


class FederationProvider:
    def __init__(self):
        pass

    def get_token(self):
        raise NotImplemented()


class AzureSystemAssignedManagedIdentityFederationProvider(FederationProvider):
    def __init__(self, audience: str = None):
        self.audience = audience if audience else 'https://management.azure.com/'
        super().__init__()

    def get_token(self):
        # azure-identity is not a hard requirement of this SDK but is required for the
        # azure provider so checking to ensure it exists
        try:
            from azure.identity._exceptions import CredentialUnavailableError
        except ImportError:
            raise Exception('azure-identity required - please install azure-identity package to use the azure managed '
                            'identity federation provider')

        try:
            from azure.identity import ManagedIdentityCredential
            token = ManagedIdentityCredential().get_token(self.audience).token
            return f'OIDC::{token}'
        except ImportError:
            raise Exception('azure-identity required - please install azure-identity package to use the azure managed '
                            'identity federation provider')
        except CredentialUnavailableError:
            msg = 'the codebase is not executing in a azure environment or some other issue is causing the ' \
                  'managed identity credentials to be unavailable'
            raise exceptions.NotExecutingInAzureEnvironment(msg)


class AzureUserAssignedManagedIdentityFederationProvider(FederationProvider):
    def __init__(self, client_id: str, audience: str = None):
        self.audience = audience if audience else 'https://management.azure.com/'
        self.client_id = client_id
        super().__init__()

    def get_token(self):
        # azure-identity is not a hard requirement of this SDK but is required for the
        # azure provider so checking to ensure it exists
        try:
            from azure.identity._exceptions import CredentialUnavailableError
        except ImportError:
            raise Exception('azure-identity required - please install azure-identity package to use the azure managed '
                            'identity federation provider')

        try:
            from azure.identity import ManagedIdentityCredential
            token = ManagedIdentityCredential(client_id=self.client_id).get_token(self.audience).token
            return f'OIDC::{token}'
        except ImportError:
            raise Exception('azure-identity required - please install azure-identity package to use the azure managed '
                            'identity federation provider')
        except CredentialUnavailableError:
            msg = 'the codebase is not executing in a azure environment or some other issue is causing the ' \
                  'managed identity credentials to be unavailable'
            raise exceptions.NotExecutingInAzureEnvironment(msg)


class GithubFederationProvider(FederationProvider):
    def __init__(self, audience: str = None):
        self.audience = audience
        super().__init__()

    def get_token(self):
        import requests
        url = os.environ.get('ACTIONS_ID_TOKEN_REQUEST_URL')
        bearer_token = os.environ.get('ACTIONS_ID_TOKEN_REQUEST_TOKEN')

        if not url or not bearer_token:
            msg = 'the codebase is not executing in a github environment and/or the action is ' \
                  'is not set to use oidc permissions'
            raise exceptions.NotExecutingInGithubEnvironment(msg)

        headers = {
            'User-Agent': 'actions/oidc-client',
            'Authorization': f'Bearer {bearer_token}'
        }

        if self.audience:
            url += f'&audience={self.audience}'

        response = requests.get(url, headers=headers)
        return f'OIDC::{response.json()["value"]}'


class AwsFederationProvider(FederationProvider):
    def __init__(self, profile: str, tenant: str, duration: int = 900):
        from ..britive import Britive  # doing import here to avoid circular dependency
        self.profile = profile
        self.duration = duration
        temp_tenant = tenant or os.getenv('BRITIVE_TENANT')
        if not temp_tenant:
            print('Error: the aws federation provider requires the britive tenant as part of the signing algorithm')
            raise exceptions.TenantMissingError()
        self.tenant = Britive.parse_tenant(temp_tenant).split(':')[0]  # remove the port if it exists
        super().__init__()

    @staticmethod
    def sign(key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    @staticmethod
    def get_signature_key(key, date_stamp, region_name, service_name):
        k_date = AwsFederationProvider.sign(('AWS4' + key).encode('utf-8'), date_stamp)
        k_region = AwsFederationProvider.sign(k_date, region_name)
        k_service = AwsFederationProvider.sign(k_region, service_name)
        k_signing = AwsFederationProvider.sign(k_service, 'aws4_request')
        return k_signing

    def get_token(self):
        # boto3 is not a hard requirement of this SDK but is required for the
        # aws provider so checking to ensure it exists
        try:
            import boto3
            import botocore.exceptions as botoexceptions
        except ImportError:
            raise Exception('boto3 required - please install boto3 package to use the aws federation provider')

        # and do all the complex logic for sigv4 signing the sts get-caller-identity endpoint
        session = None
        try:
            session = boto3.Session(profile_name=self.profile)
        except botoexceptions.ProfileNotFound as e:
            raise Exception(f'Error: {str(e)}')

        creds = session.get_credentials()
        access_key_id = creds.access_key
        secret_access_key = creds.secret_key
        session_token = creds.token

        # set some defaults
        service = 'sts'
        region = session.region_name or 'us-east-1'  # default to us-east-1 if no region available
        endpoint = f'https://sts.{region}.amazonaws.com/'
        request_body = 'Action=GetCallerIdentity&Version=2011-06-15'
        t = datetime.datetime.utcnow()
        date_stamp = t.strftime('%Y%m%d')  # Date w/o time, used in credential scope

        # these are all possible headers for 1) canonical/signed 2) actual request
        headers = {
            'host': {
                'value': f'sts.{region}.amazonaws.com',
                'include_in_request': False,  # britive backend will auto-add host header when calling sts endpoint
            },
            'x-amz-date': {
                'value': t.strftime('%Y%m%dT%H%M%SZ'),
                'include_in_request': True
            },
            'x-britive-workload-aws-tenant': {
                'value': self.tenant,
                'include_in_request': True
            },
            'x-britive-expires': {
                'value': f'{(t + datetime.timedelta(seconds=self.duration)).strftime("%Y%m%dT%H%M%SZ")}',
                'include_in_request': True
            }
        }

        # we only want to include the session token if we are dealing with temporary credentials
        # otherwise the sigv4 signing method does not need this value (which would be None)
        if session_token:
            headers['x-amz-security-token'] = {
                'value': session_token,
                'include_in_request': True
            }

        # ************* TASK 1: CREATE A CANONICAL REQUEST *************
        # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html

        # Step 1: Create the canonical headers. Header names must be trimmed and lowercase, and sorted in code point
        # order from low to high. Note that there is a trailing \n.
        canonical_header_parts = sorted([f'{h.strip().lower()}:{v["value"]}' for h, v in headers.items()])
        canonical_headers = '\n'.join(canonical_header_parts) + '\n'

        # Step 2: Create the list of signed headers. This lists the headers in the canonical_headers list,
        # delimited with ";" and in alpha order. Note: The request can include any headers; canonical_headers and
        # signed_headers include those that you want to be included in the hash of the request.
        signed_headers = ';'.join(sorted([h.strip().lower() for h, v in headers.items()]))

        # Step 3: Create payload hash. In this example, the payload (body of the request) contains the request parameters.
        payload_hash = hashlib.sha256(request_body.encode('utf-8')).hexdigest()

        # Step 4: Combine elements to create canonical request
        canonical_request_components = [
            'POST',  # method
            '/',  # canonical_uri
            '',  # canonical_querystring
            canonical_headers,
            signed_headers,
            payload_hash
        ]
        canonical_request = '\n'.join(canonical_request_components)

        # ************* TASK 2: CREATE THE STRING TO SIGN*************
        # Match the algorithm to the hashing algorithm you use, either SHA-1 or SHA-256 (recommended)
        algorithm = 'AWS4-HMAC-SHA256'
        credential_scope_components = [
            date_stamp,
            region,
            service,
            'aws4_request'
        ]
        credential_scope = '/'.join(credential_scope_components)
        string_components_to_sign = [
            algorithm,
            headers['x-amz-date']['value'],
            credential_scope,
            hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
        ]
        string_to_sign = '\n'.join(string_components_to_sign)

        # ************* TASK 3: CALCULATE THE SIGNATURE *************
        # Create the signing key using the function defined above.
        signing_key = self.get_signature_key(secret_access_key, date_stamp, region, service)

        # Sign the string_to_sign using the signing_key
        signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

        # ************* TASK 4: ADD SIGNING INFORMATION TO THE REQUEST *************
        # Put the signature information in a header named Authorization.
        authorization_header_parts = [
            algorithm,
            f'Credential={access_key_id}'
        ]
        authorization_header = f'{algorithm} Credential={access_key_id}/{credential_scope}, ' \
                               f'SignedHeaders={signed_headers}, Signature={signature}'

        # Except for the authorization header, the headers must be included in the canonical_headers and
        # signed_headers values, as noted earlier. order here is not significant.
        # Python note: The 'host' header is added automatically by the Python 'requests' library.
        request_headers = {h: v['value'] for h, v in headers.items() if v['include_in_request'] and v['value']}
        request_headers['authorization'] = authorization_header

        token = {
            'iam_request_url': endpoint,
            'iam_request_body': request_body,
            'iam_request_headers': request_headers
        }

        token_encoded = base64.urlsafe_b64encode(json.dumps(token).encode('utf-8'))
        return f'AWS::{token_encoded.decode("utf-8")}'


class BitbucketFederationProvider(FederationProvider):
    def __init__(self):
        super().__init__()

    # https://support.atlassian.com/bitbucket-cloud/docs/integrate-pipelines-with-resource-servers-using-oidc/
    def get_token(self):
        id_token = os.environ.get('BITBUCKET_STEP_OIDC_TOKEN')
        if not id_token:
            msg = 'the codebase is not executing in a bitbucket environment and/or the `oidc` flag ' \
                  'is not set on the pipeline step'
            raise exceptions.NotExecutingInBitbucketEnvironment(msg)
        return f'OIDC::{id_token}'


class SpaceliftFederationProvider(FederationProvider):
    def __init__(self):
        super().__init__()

    # https://docs.spacelift.io/integrations/cloud-providers/oidc/
    def get_token(self):
        id_token = os.environ.get('SPACELIFT_OIDC_TOKEN')
        if not id_token:
            msg = 'the codebase is not executing in a spacelift.io environment or not using a paid account'
            raise exceptions.NotExecutingInSpaceliftEnvironment(msg)
        return f'OIDC::{id_token}'




class Scans:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def scan(self, application_id: str, environment_id: str = None) -> dict:
        """
        Initiate a scan of the provided application and optionally provided environment.

        `environment_id` is only required only for applications which have
        `catalogApplication.supportsEnvironmentScanning` set to `False`.

        Scans are asynchronous operations. The response will include a `taskId` which can be used to make calls
        to `Scans.status()` to obtain the current status of the scan.

        Note that scans can also be initiated from the Applications and Environments classes. The same type of
        scan will be performed no matter where it is initiated.

        :param application_id: The ID of the application to scan.
        :param environment_id: Optionally the ID of the environment to scan.
        :return: Details of the scan that was initiated.
        """

        if environment_id:
            return self.britive.post(f'{self.base_url}/{application_id}/environments/{environment_id}/scans')
        else:
            return self.britive.post(f'{self.base_url}/{application_id}/scan')

    def status(self, task_id: str) -> dict:
        """
        Provide details on the current status of a scan.

        :param task_id: The ID of the task which was obtained when a scan was initiated.
        :return: Details on the current status of the scan.
        """

        return self.britive.get(f'{self.base_url}/tasks/{task_id}/status')

    def history(self, application_id: str, filter_expression: str = None) -> list:
        """
        Return a list of all scans in the past 90 days for the application ID provided.

        :param application_id: The ID of the application for which scan history will be retrieved.
        :param filter_expression: Optional filter expression to filter the results server side. Can filter based
            on `name`, `status`, and `integrity checks`. The supported operators are `eq` and `co`.
            Example: `name co "Dev Account"`.
        :return: A list of historical scans.
        """

        params = {
            'page': 0,
            'size': 100
        }
        if filter_expression:
            params['filter'] = filter_expression
        return self.britive.get(f'{self.base_url}/{application_id}/scans/env-status/history', params=params)

    def diff(self, resource: str, application_id: str, environment_id: str = None) -> list:
        """
        Retrieve changes between the last two consecutive scans.

        :param resource: valid options are `permissions`, `groups`, and `accounts`.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :return: List of changes between the last two consecutive scans of the given resource.
        """

        if resource not in ['permissions', 'groups', 'accounts']:
            raise ValueError(f'invalid resource {resource}')

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/{resource}/memberships-last-scan-delta'
        return self.britive.get(url)



import json
from . import exceptions
from typing import Union


creation_defaults = {
    'expirationDuration': 3600000,
    'extensionDuration': 1800000,
    'notificationPriorToExpiration': 300000,
    'extendable': False,
    'extensionLimit': '1',
    'status': 'active',
    'destinationUrl': '',
    'useDefaultAppUrl': True,
    'description': ''
}

update_fields_to_keep: list = list(creation_defaults)
update_fields_to_keep.append('name')
update_fields_to_keep.remove('status')


class Profiles:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'
        self.permissions = ProfilePermissions(britive)
        self.session_attributes = ProfileSessionAttributes(britive)
        self.policies = ProfilePolicies(britive)

    def create(self, application_id: str, name: str, **kwargs) -> dict:
        """
        Create a profile.

        :param application_id: The ID of the application under which the profile will be created.
        :param name: The name of the profile. This is the only required argument.
        :param kwargs: A key/value mapping consisting of the following fields. If any/all are omitted default values
            will be used. The keys and default values are provided below.

            - expirationDuration: 3600000
            - extensionDuration: 1800000
            - notificationPriorToExpiration: 300000
            - extendable: False
            - extensionLimit: '1'
            - status: 'active'
            - destinationUrl: ''
            - useDefaultAppUrl: True
            - description: ''
            - scope: if not provided, no scopes will be applied. If provided it must follow the
                format listed below.

                [
                    {
                        'type': 'EnvironmentGroup'|'Environment',
                        'value':'ID'
                    },
                ]

        :return: Details of the newly created profile.
        """

        kwargs['appContainerId'] = application_id
        kwargs['name'] = name  # required field, so it is being called out explicitly in the method parameters

        # merge defaults and provided information - keys in kwargs will overwrite the defaults in creation_defaults
        data = {**creation_defaults, **kwargs}  # note python 3.5 or greater but only 3.5 and up are supported so okay!

        return self.britive.post(f'{self.base_url}/{application_id}/paps', json=data)

    def list(self, application_id: str, filter_expression: str = None, environment_association: str = None,
             include_policies: bool = False) -> list:
        """
        Return an optionally filtered list of profiles associated with the specified application.

        :param application_id: The ID of the application.
        :param filter_expression: Can filter based on `name`, `status`, `integrity check`. Valid operators are `eq` and
            `co`. Example: name co "Dev Account"
        :param environment_association: Only list profiles with associations to the specified environment. Cannot be
            used in conjunction with `filter_expression`. Example: `environment_association="109876543210"`
        :param include_policies: Defaults to False. If set to True will include all policies on each profile and all
            members on each policy. Cannot be used in conjunction with `filter_expression`.
        :return: List of profiles.
        """

        if filter_expression and environment_association:
            raise exceptions.InvalidRequest(
                'Cannot specify `filter_expression` and `environment_association` in the same request.'
            )

        params = {
            'page': 0,
            'size': 100,
            'view': 'summary'  # this is required - omitting it results in a 400 not authorized error
        }

        if include_policies:
            params['view'] = 'includePolicies'

        if filter_expression:
            params['filter'] = filter_expression

        if environment_association:
            params['environment'] = environment_association

        return self.britive.get(f'{self.base_url}/{application_id}/paps', params=params)

    def get(self, application_id: str, profile_id: str, summary: bool = None) -> dict:
        """
        Return details of the provided profile.

        :param application_id: The ID of the application.
        :param profile_id: The ID of the profile.
        :param summary: Whether to provide a summarized response. Defaults to None to support backwards compatibility
            with the legacy functionality/way of obtaining details of the profile. Setting to True will return a
            summarized set of attributes for the profile. Setting to False will return a larger set of attributes
            for the profile.
        :return: Details of the profile.
        :raises: ProfileNotFound if the profile does not exist.
        """

        if summary is None:
            for profile in self.list(application_id=application_id):
                if profile['papId'] == profile_id:
                    return profile

            raise exceptions.ProfileNotFound()
        else:
            params = {}
            if summary:
                params['view'] = 'summary'
            return self.britive.get(f'{self.britive.base_url}/paps/{profile_id}', params=params)

    def update(self, application_id: str, profile_id: str, **kwargs) -> dict:
        """
        Update details of the specified profile.

        :param application_id: The ID of the application.
        :param profile_id: The ID of the profile to update.
        :param kwargs: Refer to the `create()` method for details on parameters that can be provided. For this update
            action no default values will be injected for missing parameters.
        :return: Details of the updated profile.
        """

        existing = self.get(application_id=application_id, profile_id=profile_id, summary=True)
        base = {key: existing[key] for key in update_fields_to_keep}

        kwargs['appContainerId'] = application_id
        data = {**base, **kwargs}

        return self.britive.patch(f'{self.base_url}/{application_id}/paps/{profile_id}', json=data)

    def available_resources(self, profile_id: str, filter_expression: str = None) -> list:
        """
        AZURE ONLY. This API is applicable to Azure applications only.

        Returns the list of all resource scopes that are available and can be added to a profile.

        :param profile_id: The ID of the profile.
        :param filter_expression: Can filter based on `name`, `status`, `integrity check`. Valid operators are `eq` and
            `co`. Example: name co "Dev Account"
        :return: List of available resource scopes.
        """

        params = {
            'page': 0,
            'size': 100
        }

        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(f'{self.britive.base_url}/paps/{profile_id}/resources', params=params)

    def get_scopes(self, profile_id: str) -> list:
        """
        Get the scopes associated with the specified profile.

        :param profile_id: The ID of the profile for which scopes will be updated.
        :return: List of scopes and associated details.
        """

        return self.britive.get(f'{self.britive.base_url}/paps/{profile_id}/scopes')

    def set_scopes(self, profile_id: str, scopes: list) -> list:
        """
        Update the scopes associated with the specified profile.

        :param profile_id: The ID of the profile for which scopes will be updated.
        :param scopes: List of scopes. Example  below.
            [
                {
                    'type': 'EnvironmentGroup'|'Environment',
                    'value':'ID'
                },
            ]
        :return: List of scopes and associated details.
        """

        return self.britive.post(f'{self.britive.base_url}/paps/{profile_id}/scopes', json=scopes)

    def add_single_environment_scope(self, profile_id: str, environment_id: str) -> None:
        """
        Add a single environment to the scopes associated with the specified profile.

        This API call is useful when there are a large number of environments (500+) as the
        `set_scopes()` call may time out.

        :param profile_id: The ID of the profile.
        :param environment_id: The ID of the environment which will be added to the profile scopes.
        :return: None
        """

        return self.britive.patch(f'{self.britive.base_url}/paps/{profile_id}/scopes/{environment_id}')

    def remove_single_environment_scope(self, profile_id: str, environment_id: str) -> None:
        """
        Remove a single environment from the scopes associated with the specified profile.

        This API call is useful when there are a large number of environments (500+) as the
        `set_scopes()` call may time out.

        :param profile_id: The ID of the profile.
        :param environment_id: The ID of the environment which will be removed from the profile scopes.
        :return: None
        """

        return self.britive.delete(f'{self.britive.base_url}/paps/{profile_id}/scopes/{environment_id}')

    def enable(self, application_id: str, profile_id: str) -> dict:
        """
        Enables a profile.

        :param application_id: The ID of the application.
        :param profile_id: The ID of the profile to enable.
        :return: Details of the enabled profile.
        """

        return self.britive.post(f'{self.base_url}/{application_id}/paps/{profile_id}/enabled-statuses')

    def disable(self, application_id: str, profile_id: str) -> dict:
        """
        Disables a profile.

        :param application_id: The ID of the application.
        :param profile_id: The ID of the profile to disable.
        :return: Details of the disabled profile.
        """

        return self.britive.post(f'{self.base_url}/{application_id}/paps/{profile_id}/disabled-statuses')

    def delete(self, application_id: str, profile_id: str) -> None:
        """
        Deletes a profile.

        :param application_id: The ID of the application.
        :param profile_id: The ID of the profile to delete.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{application_id}/paps/{profile_id}')


class ProfilePermissions:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/paps'
        self.constraints = ProfilePermissionConstraints(britive)

    def add(self, profile_id: str, permission_type: str, permission_name: str) -> dict:
        """
        Add a permission to a profile.

        Call `list_available()` to see what permissions can be added.

        Note that for AWS and OCI permissions are not assigned to profiles as the permissions are tied into the
        cloud provider directly (AssumeRole for AWS).

        :param profile_id: The ID of the profile.
        :param permission_type: The type of permission. Valid values are `role`, `group`, and `policy`.
        :param permission_name: The name of the permission.
        :return: Details of the permission added.
        """

        data = {
            'op': 'add',
            'permission': {
                'name': permission_name,
                'type': permission_type
            }
        }

        return self.britive.post(f'{self.base_url}/{profile_id}/permissions', json=data)

    def list_assigned(self, profile_id: str, filter_expression: str = None) -> list:
        """
        List the permissions assigned to the profile.

        :param profile_id: The ID of the profile.
        :param filter_expression: Can filter based on `name`, `status`, `integrity check`. Valid operators are `eq` and
            `co`. Example: name co "Dev Account"
        :return: List of permissions assigned to the profile.
        """

        params = {
            'page': 0,
            'size': 100
        }

        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(f'{self.base_url}/{profile_id}/permissions', params=params)

    def list_available(self, profile_id: str) -> list:
        """
        List permissions available to be assigned to the profile.

        Note that for AWS and OCI permissions are not assigned to profiles as the permissions are tied into the
        cloud provider directly (AssumeRole for AWS).

        :param profile_id: The ID of the profile.
        :return: List of permissions that are available to be assigned to the profile.
        """

        params = {
            'page': 0,
            'size': 100,
            'query': 'available'
        }
        return self.britive.get(f'{self.base_url}/{profile_id}/permissions', params=params)

    def remove(self, profile_id: str, permission_type: str, permission_name: str) -> dict:
        """
        Remove a permission to a profile.

        :param profile_id: The ID of the profile.
        :param permission_type: The type of permission. Valid values are `role`, `group`, and `policy`.
        :param permission_name: The name of the permission.
        :return: Details of the permission removed.
        """

        data = {
            'op': 'remove',
            'permission': {
                'name': permission_name,
                'type': permission_type
            }
        }

        return self.britive.post(f'{self.base_url}/{profile_id}/permissions', json=data)


class ProfilePermissionConstraints:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/paps'

    def list_supported_types(self, profile_id: str, permission_name: str, permission_type: str = 'role') -> list:
        """
        Lists the supported constraint types.

        :param profile_id: The ID of the profile.
        :param permission_name: The name of the permission for which to list supported constraints.
        :param permission_type: The type of permission. Defaults to `role`.
        :returns: List of supported constraint types.
        """

        url = f'{self.base_url}/{profile_id}/permissions/{permission_name}/' \
              f'{permission_type}/supported-constraint-types'
        return self.britive.get(url)

    def get(self, profile_id: str, permission_name: str, constraint_type: str, permission_type: str = 'role') -> list:
        """
        Gets the list of constraints.

        :param profile_id: The ID of the profile.
        :param permission_name: The name of the permission for which to list supported constraints.
        :param constraint_type: The type of constraint.
        :param permission_type: The type of permission. Defaults to `role`.
        :returns: List of constraints for the given constraint type.
        """

        url = f'{self.base_url}/{profile_id}/permissions/{permission_name}/' \
              f'{permission_type}/constraints/{constraint_type}'
        return self.britive.get(url).get('result')

    def lint_condition(self, profile_id: str, permission_name: str, expression: str,
                       permission_type: str = 'role') -> dict:
        """
        Lint the provided condition expression.

        :param profile_id: The ID of the profile.
        :param permission_name: The name of the permission for which to lint the condition expression.
        :param expression: The condition expression to lint.
        :param permission_type: The type of permission. Defaults to `role`.
        :returns: Results of the lint operation.
        """

        url = f'{self.base_url}/{profile_id}/permissions/{permission_name}/' \
              f'{permission_type}/constraints/condition'

        params = {
            'operation': 'validate'
        }

        data = {
            'expression': expression
        }

        return self.britive.put(url, params=params, json=data)

    def add(self, profile_id: str, permission_name: str, constraint_type: str, constraint: dict,
            permission_type: str = 'role') -> None:
        """
        Adds the given constraint.

        :param profile_id: The ID of the profile.
        :param permission_name: The name of the permission for which the constraint should be added.
        :param constraint_type: The type of constraint.
        :param constraint: The constraint to add. If `constraint_type == 'condition'` then this parameter should be a
            dict with fields `title`, `description`, and `expression`. Otherwise, this parameter should be a dict with
            field `name` or string value.
        :param permission_type: The type of permission. Defaults to `role`.
        :returns: None.
        """

        url = f'{self.base_url}/{profile_id}/permissions/{permission_name}/' \
              f'{permission_type}/constraints/{constraint_type}'

        params = {
            'operation': 'add'
        }

        return self.britive.put(url, params=params, json=constraint)

    def remove(self, profile_id: str, permission_name: str, constraint_type: str, constraint: dict = None,
               permission_type: str = 'role') -> None:
        """
        Removes the given constraint.

        :param profile_id: The ID of the profile.
        :param permission_name: The name of the permission for which the constraint should be removed.
        :param constraint_type: The type of constraint.
        :param constraint: The constraint to remove. If `constraint_type == 'condition'` then omit this parameter or
            set it to `None`. Otherwise, this parameter should be a dict with field `name` or string value.
        :param permission_type: The type of permission. Defaults to `role`.
        :returns: None.
        """

        url = f'{self.base_url}/{profile_id}/permissions/{permission_name}/' \
              f'{permission_type}/constraints/{constraint_type}'
        params = {
            'operation': 'remove'
        }
        if constraint is None:
            constraint = {}

        return self.britive.put(url, params=params, json=constraint)


class ProfileSessionAttributes:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/paps'

    def add_static(self, profile_id: str, tag_name: str, tag_value: str, transitive: bool = False) -> dict:
        """
        AWS ONLY - Add a static session attribute to the profile.

        :param profile_id: The ID of the profile.
        :param tag_name: The name of the session tag to include in the AssumeRoleWithSAML call.
        :param tag_value: THe value of the session tag to include in the AssumeRoleWithSAML call.
        :param transitive: Set to True to mark the session tag as transitive. Review AWS documentation on why you
            may or may not want this.
        :return: Details of added attribute.
        """

        data = {
            'sessionAttributeType': 'Static',
            'transitive': transitive,
            'attributeSchemaId': None,
            'mappingName': tag_name,
            'attributeValue': tag_value
        }

        return self.britive.post(f'{self.base_url}/{profile_id}/session-attributes', json=data)

    def add_dynamic(self, profile_id: str, identity_attribute_id: str, tag_name: str,
                    transitive: bool = False) -> dict:
        """
        AWS ONLY - Add a dynamic session attribute to the profile.

        The value will be sourced from the identity attribute specified.

        :param profile_id: The ID of the profile.
        :param identity_attribute_id: The ID of the identity attribute.  Call `britive.identity_attributes.list()`
            for details on which attributes can be provided.
        :param tag_name: The name of the session tag to include in the AssumeRoleWithSAML call. The value will be
            dynamically determined based on the value of the specified identity attribute.
        :param transitive: Set to True to mark the session tag as transitive. Review AWS documentation on why you
            may or may not want this.
        :return: Details of added attribute.
        """

        data = {
            'sessionAttributeType': 'Identity',
            'transitive': transitive,
            'attributeSchemaId': identity_attribute_id,
            'mappingName': tag_name,
            'attributeValue': None
        }

        return self.britive.post(f'{self.base_url}/{profile_id}/session-attributes', json=data)

    def update_static(self, profile_id: str, attribute_id, tag_name: str, tag_value: str,
                      transitive: bool = False) -> None:
        """
        AWS ONLY - Update the static session attribute to the profile.

        :param profile_id: The ID of the profile.
        :param attribute_id: The ID of the session attribute to update.
        :param tag_name: The name of the session tag to include in the AssumeRoleWithSAML call.
        :param tag_value: THe value of the session tag to include in the AssumeRoleWithSAML call.
        :param transitive: Set to True to mark the session tag as transitive. Review AWS documentation on why you
            may or may not want this.
        :return: None.
        """

        data = {
            'sessionAttributeType': 'Static',
            'transitive': transitive,
            'attributeSchemaId': None,
            'mappingName': tag_name,
            'attributeValue': tag_value,
            'id': attribute_id
        }

        return self.britive.put(f'{self.base_url}/{profile_id}/session-attributes', json=data)

    def update_dynamic(self, profile_id: str, attribute_id: str, identity_attribute_id: str, tag_name: str,
                       transitive: bool = False) -> dict:
        """
        AWS ONLY - Update the dynamic session attribute to the profile.

        :param profile_id: The ID of the profile.
        :param attribute_id: The ID of the session attribute to update.
        :param identity_attribute_id: The ID of the identity attribute.  Call `britive.identity_attributes.list()`
            for details on which attributes can be provided.
        :param tag_name: The name of the session tag to include in the AssumeRoleWithSAML call. The value will be
            dynamically determined based on the value of the specified identity attribute.
        :param transitive: Set to True to mark the session tag as transitive. Review AWS documentation on why you
            may or may not want this.
        :return: Details of added attribute.
        """

        data = {
            'sessionAttributeType': 'Identity',
            'transitive': transitive,
            'attributeSchemaId': identity_attribute_id,
            'mappingName': tag_name,
            'attributeValue': None,
            'id': attribute_id
        }

        return self.britive.put(f'{self.base_url}/{profile_id}/session-attributes', json=data)

    def list(self, profile_id: str) -> list:
        """
        Return a list of session attributes associated with the profile.

        :param profile_id: The ID of the profile.
        :return: List of session attributes associated with the profile.
        """

        return self.britive.get(f'{self.base_url}/{profile_id}/session-attributes')

    def remove(self, profile_id: str, attribute_id: str) -> None:
        """
        Remove an attribute from the profile.

        :param profile_id: The ID of the profile.
        :param attribute_id: The ID of the session attribute.
        :return: None.
        """

        return self.britive.delete(f'{self.base_url}/{profile_id}/session-attributes/{attribute_id}')


class ProfilePolicies:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/paps'

    def build(self, name: str, description: str = '', draft: bool = False, active: bool = True,
              read_only: bool = False, users: list = None, tags: list = None, service_identities: list = None,
              ips: list = None, from_time: str = None, to_time: str = None, date_schedule: dict = None,
              days_schedule: dict = None, approval_notification_medium: Union[str, list] = None,
              time_to_approve: int = 5, access_validity_time: int = 120, approver_users: list = None,
              approver_tags: list = None, access_type: str = 'Allow', identifier_type: str = 'name',
              condition_as_dict: bool = False) -> dict:
        """
        Build a policy document given the provided inputs.

        :param name: The name of the policy.
        :param description: An optional description of the policy.
        :param draft: Indicates if the policy is a draft. Defaults to `False`.
        :param active: Indicates if the policy is active. Defaults to `True`.
        :param read_only: Indicates if the policy is a read only. Defaults to `False`.
        :param users: Optional list of user names or ids to which this policy applies.
        :param tags: Optional list of tag names or ids to which this policy applies.
        :param service_identities: Optional list of service identity names or ids to which this policy applies.
        :param ips: Optional list of IP addresses for which this policy applies. Provide in CIDR notation
            or dotted decimal format for individual (/32) IP addresses.
        :param from_time: The start date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the start date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring start time. If this parameter is
            provided then `to_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param to_time: The end date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the end date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring end time. If this parameter is
            provided then `from_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param date_schedule: A dict in the format
            {
                'fromDate': '2022-10-29 10:30:00',
                'toDate': '2022-11-05 18:30:00',
                'timezone': 'UTC'
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param days_schedule: A dict in the format
            {
                'fromTime': '10:30:00',
                'toTime': '18:30:00',
                'timezone': 'UTC',
                'days': ['MONDAY','TUESDAY','WEDNESDAY','THURSDAY','FRIDAY','SATURDAY','SUNDAY']
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param approval_notification_medium: Optional notification medium name to which approval requests will be
            delivered. Can also specify a list of notification medium names. Specifying this parameter indicates the
            desire to enable approvals for this policy.
        :param time_to_approve: Optional number of minutes to wait for an approval before denying the action. Defaults
            to 5 minutes.
        :param access_validity_time: Optional number of minutes the access is valid after approval. Defaults to 120
            minutes.
        :param approver_users: Optional list of user names or ids who are to be considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param approver_tags: Optional list of tag names who are considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param access_type: The type of access this policy provides. Valid values are `Allow` and `Deny`. Defaults
            to `Allow`.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers are being provided to the other parameters. Either all identifiers must be names or all
            identifiers must be IDs.
        :param condition_as_dict: Prior to version 2.22.0 the only acceptable format for the condition block of
            a policy was as a stringifed json object. As of 2.22.0 the condition block can also be built as a raw
            python dictionary. This parameter will default to `False` to support backwards compatibility. Setting to
            `True` will result in the policy condition being returned/built as a python dictionary.
        :return: A dict which can be provided as a profile policy to `create` and `update`.
        """

        policy = self.britive.system.policies.build(
            name=name,
            description=description,
            draft=draft,
            active=active,
            read_only=read_only,
            users=users,
            tags=tags,
            service_identities=service_identities,
            ips=ips,
            from_time=from_time,
            to_time=to_time,
            date_schedule=date_schedule,
            days_schedule=days_schedule,
            approval_notification_medium=approval_notification_medium,
            time_to_approve=time_to_approve,
            access_validity_time=access_validity_time,
            approver_users=approver_users,
            approver_tags=approver_tags,
            access_type=access_type,
            identifier_type=identifier_type,
            condition_as_dict=condition_as_dict
        )

        # clean up the generic policy response and customize for profiles
        policy.pop('permissions', None)
        policy.pop('roles', None)
        policy['consumer'] = 'papservice'

        return policy

    def list(self, profile_id: str) -> list:
        """
        List all policies associated with the provided profile.

        Only applicable to tenants using version 2 of profiles. If the tenant is on version 1 of profiles then use
        `britive.profiles.tags.*` and `britive.profiles.identities.*` instead.

        :param profile_id: The ID of the profile.
        :return: List of policies.
        """

        return self.britive.get(f'{self.base_url}/{profile_id}/policies')

    def get(self, profile_id: str, policy_id: str, condition_as_dict: bool = False) -> dict:
        """
        Retrieve details about a specific policy which is associated with the provided profile.

        Only applicable to tenants using version 2 of profiles. If the tenant is on version 1 of profiles then use
        `britive.profiles.tags.*` and `britive.profiles.identities.*` instead.

        :param profile_id: The ID of the profile.
        :param policy_id: The ID of the policy.
        :param condition_as_dict: Prior to version 2.22.0 a policy condition block was always returned as stringified
            json. As of 2.22.0 the SDK now supports returning the condition block of a policy as either stringified json
            or a raw python dictionary. The Britive backend will also return the condition block in either format,
            depending on a query parameter value. Setting this value to `True` will result in the condition block being
            returned as a python dictionary. The default of `False` is to support backwards compatibility.
        :return: Details of the policy.
        """

        params = {
            'conditionJson': condition_as_dict
        }

        policy = self.britive.get(f'{self.base_url}/{profile_id}/policies/{policy_id}', params=params)

        # it seems profile policy is not honoring conditionJson parameter so doing some extra work here
        # to get things into the correct format. if in the future that changes we can perhaps remove
        # the below logic.
        if 'condition' in policy and condition_as_dict and isinstance(policy['condition'], str):
            policy['condition'] = json.loads(policy['condition'])

        return policy

    def create(self, profile_id: str, policy: dict) -> dict:
        """
        Create a policy associated with the provided profile.

        Only applicable to tenants using version 2 of profiles. If the tenant is on version 1 of profiles then use
        `britive.profiles.tags.*` and `britive.profiles.identities.*` instead.

        :param profile_id: The ID of the profile.
        :param policy: The policy contents to create.
        :return: Details of the newly created policy.
        """

        return self.britive.post(f'{self.base_url}/{profile_id}/policies', json=policy)

    def update(self, profile_id: str, policy_id: str, policy: dict) -> dict:
        """
        Update the contents of the provided policy associated with the provided profile.

        Only applicable to tenants using version 2 of profiles. If the tenant is on version 1 of profiles then use
        `britive.profiles.tags.*` and `britive.profiles.identities.*` instead.

        :param profile_id: The ID of the profile.
        :param policy_id: The ID of the policy.
        :param policy: The policy to update.
        :return: Details of the updated policy.
        """

        return self.britive.patch(f'{self.base_url}/{profile_id}/policies/{policy_id}', json=policy)

    def delete(self, profile_id: str, policy_id: str) -> None:
        """
        Delete the provided policy associated with the provided profile.

        Only applicable to tenants using version 2 of profiles. If the tenant is on version 1 of profiles then use
        `britive.profiles.tags.*` and `britive.profiles.identities.*` instead.

        :param profile_id: The ID of the profile.
        :param policy_id: The ID of the policy.
        :return: None.
        """

        return self.britive.delete(f'{self.base_url}/{profile_id}/policies/{policy_id}')



from .cache import *  # will also import some globals like `britive`
from britive import exceptions


def test_create(cached_identity_provider):
    assert isinstance(cached_identity_provider, dict)


def test_list(cached_identity_provider):
    idps = britive.identity_providers.list()
    assert isinstance(idps, list)
    assert len(idps) > 0
    assert isinstance(idps[0], dict)
    assert cached_identity_provider['id'] in [idp['id'] for idp in idps]


def test_get_by_id(cached_identity_provider):
    idp = britive.identity_providers.get(identity_provider_id=cached_identity_provider['id'])
    assert isinstance(idp, dict)
    assert idp['id'] == cached_identity_provider['id']


def test_get_by_name(cached_identity_provider):
    idp = britive.identity_providers.get_by_name(identity_provider_name=cached_identity_provider['name'])
    assert isinstance(idp, dict)
    assert idp['name'] == cached_identity_provider['name']


def test_update(cached_identity_provider):
    response = britive.identity_providers.update(
        identity_provider_id=cached_identity_provider['id'],
        sso_provider='Azure'
    )
    assert response is None
    idp = britive.identity_providers.get_by_name(identity_provider_name=cached_identity_provider['name'])
    assert isinstance(idp, dict)
    assert idp['ssoProvider'] == 'Azure'


def test_scim_token_create(cached_scim_token):
    assert isinstance(cached_scim_token, dict)
    assert cached_scim_token['tokenExpirationDays'] == 60


def test_scim_token_get(cached_scim_token, cached_identity_provider):
    token = britive.identity_providers.scim_tokens.get(identity_provider_id=cached_identity_provider['id'])
    assert isinstance(token, dict)
    assert token['name'] == cached_scim_token['name']


def test_scim_token_update(cached_identity_provider):
    response = britive.identity_providers.scim_tokens.update(
        identity_provider_id=cached_identity_provider['id'],
        token_expiration_days=30
    )
    assert response is None
    token = britive.identity_providers.scim_tokens.get(identity_provider_id=cached_identity_provider['id'])
    assert token['tokenExpirationDays'] == 30


def test_scim_attributes_list():
    attributes = britive.identity_providers.scim_attributes.list()
    assert isinstance(attributes, list)
    assert len(attributes) > 0
    assert isinstance(attributes[0], str)


def test_scim_tokens_update_attribute_mapping(cached_identity_provider):

    attributes = britive.identity_attributes.list()
    phone_id = None
    for attribute in attributes:
        if attribute['builtIn'] and attribute['name'] == 'Phone':
            phone_id = attribute['id']
            break

    mappings = [
        {
            'scimAttributeName': 'phoneNumbers[type eq "work"]',
            'builtIn': True,
            'attributeId': phone_id,
            'attributeName': 'Phone',
            'op': 'remove'
        }
    ]
    response = britive.identity_providers.scim_attributes.update_mapping(
        identity_provider_id=cached_identity_provider['id'],
        mappings=mappings
    )
    assert response is None
    idp = britive.identity_providers.get(identity_provider_id=cached_identity_provider['id'])
    assert 'userAttributeScimMappings' in idp.keys()
    assert isinstance(idp['userAttributeScimMappings'], list)
    mappings = idp['userAttributeScimMappings']
    assert 'Phone' not in [m['attributeName'] for m in mappings]


def test_configure_mfa(cached_identity_provider):
    with pytest.raises(exceptions.InvalidRequest) as e:
        response = britive.identity_providers.configure_mfa(
            identity_provider_id=cached_identity_provider['id'],
            root_user=False,
            non_root_user=True
        )
    assert 'E1001 - MFA can only be enabled on default identity provider' in str(e)


def test_signing_certificate():
    cert = britive.identity_providers.signing_certificate()
    assert isinstance(cert, str)
    assert '-----BEGIN CERTIFICATE-----' in cert


def test_set_metadata(cached_identity_provider):
    response = britive.identity_providers.set_metadata(
        identity_provider_id=cached_identity_provider['id'],
        metadata_xml=britive.saml.metadata()
    )
    assert isinstance(response, dict)
    assert 'certificateDn' in response.keys()
    assert response['certificateDn'] == f'CN={britive.tenant.split(".")[0]}'


def test_delete(cached_identity_provider):
    response = britive.identity_providers.delete(identity_provider_id=cached_identity_provider['id'])
    assert response is None
    cleanup('identity-provider')
    cleanup('scim-token')



from .cache import *  # will also import some globals like `britive`


def test_list():
    response = britive.system.consumers.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)
    assert 'name' in response[0].keys()

